// This file is part of WifiCC3000 - the Managed Wifi Driver for the TI CC3000
// WifiCC3000 is free software licensed under the Apache License 2.0
// by Valkyrie-MT and Ziggurat29

using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using System.Collections;
using System.Threading;
using System.Text;
using System.IO;

namespace Networking
{
    public class CC3000Driver
    {
        public bool WriteEnabled = false;

        public static uint lastMem1 = 0;
        public static uint lastMem2 = 0;

        public static uint sendCount = 0;

        public const uint MaxSockets = 4; // the TI CC3000 only supports 4 sockets. 

        internal object spiLock = new object();

        internal SPI spi;
        internal OutputPort cs;
        internal OutputPort pwren;
        internal InputPort irq;
        internal Timer irqCheck;

        private static byte[] ExpectedSimpleLinkResponse = new byte[] { 0x02, 0x00, 0xFF, 0x00 };

        internal byte[] spVersion = null;
        internal byte patchMode = 0;

        internal OutputPort analyzerTrigger = new OutputPort((Cpu.Pin)39, false);

        private AutoResetEvent Waiter = new AutoResetEvent(false);
        private AutoResetEvent nvWaiter = new AutoResetEvent(false);
        private AutoResetEvent BlockingHold = new AutoResetEvent(false);
        private AutoResetEvent ResponseReadyFlag = new AutoResetEvent(false);
        //internal ushort WaiterOpCode = (ushort)HciCommand.UNKNOWN;
        internal Hashtable WaiterOpCodes = new Hashtable();

        internal AutoResetEvent InitWaiter = new AutoResetEvent(false);
        internal bool useDHCP = true;

        public bool IsResetting = false;

        internal TimeSpan LastResponse = TimeSpan.MaxValue;
        internal Timer WatchDogTimer;

        private static int packetCount = 0;

        //private Hashtable responses = new Hashtable(5);

        private bool IsInitialized = false;
        private bool IsWriting = false;
        private bool IsReading = false;


        private bool SuppressConnectionEvent = false;

        private byte[] nvBuffer = null;

        private Thread InitThread;
        private Thread InterruptPoller;
        private Thread ResponseDispatcher;

        internal Socket[] sockets = new Socket[MaxSockets];
        internal Hashtable Listeners = new Hashtable();

        internal SafeMessageList PendingCommands = new SafeMessageList();
        internal SafeMessageList Responses = new SafeMessageList();


        public event Networking.WifiAdapter.ConnectedEventHandler OnConnectedEvent;


        /// <summary>
        /// Instantiate and Start up the wifi adapter
        /// </summary>
        /// <param name="intPin"></param>
        /// <param name="csPin"></param>
        /// <param name="enablePin"></param>
        /// <param name="spiMod"></param>
        internal CC3000Driver(Cpu.Pin intPin, Cpu.Pin csPin, Cpu.Pin enablePin, SPI.SPI_module spiMod)
        {
            //make our spi (mode 1)
            spi = new SPI(new SPI.Configuration(
                (Cpu.Pin)(-1),        // SS-pin (we will manually do it)
                false,                // SS-pin active state (low)
                0,                    // The setup time for the SS port YYY 50 uSec
                0,                    // The hold time for the SS port
                false,                // The idle state of the clock (low)
                false,                // The sampling clock edge (trailing)
                4000,                // The SPI clock rate in KHz (15 MHz is the max for the cc3000 and 1 MHz is the recommended Minimum) http://e2e.ti.com/support/low_power_rf/f/851/t/274780.aspx?pi267162=6
                spiMod));    // The used SPI bus (refers to a MOSI MISO and SCLK pinset)

            //make our (manually controlled) cs (active low)
            cs = new OutputPort(csPin, true);

            //make our power control (active high)
            pwren = new OutputPort(enablePin, false);
            //make our interrupt input (active low)    //XXX holding off on this until I understand the device better

            // Note: Using an InterruptPort for the IRQ was causing a memory leak, so I had to switch to polling an InputPort instead.  :(
            irq = new InputPort(intPin, true, Port.ResistorMode.Disabled);  // should this be pullup?  
            //irq = new InputPort(intPin);

            InterruptPoller = new Thread(PollInterrupt);
            InterruptPoller.Start();

            ResponseDispatcher = new Thread(PollResponses);
            ResponseDispatcher.Start();

            //irqCheck = new Timer(new TimerCallback(irq_OnInterrupt), null, 0, 50);

            //irq.OnInterrupt += irq_OnInterrupt;

            //Debug.Print("IRQ state = " + irq.Read());
        }

        
        /// <summary>
        /// 
        /// </summary>
        internal void PowerUp()
        {
            Debug.EnableGCMessages(false);

            // Note: this Start method could not be in the constructor because event attachment must happen after construction but before init!  

            //Debug.Print("Interrupt state before start: " + irq.Read());

            if (pwren.Read()) Shutdown();

            ClearNetworkProperties();

            Debug.Print("Starting up CC3000...");

            IsInitialized = false;  // Power is off, so the device is not initialized

            cs.Write(true);

            //power up
            pwren.Write(true);

            // Wait 1000ms after power enabled before communicating?  http://www.nerdkits.com/forum/thread/2462/
            //Debug.Print("Interrupt state after start: " + irq.Read());

            //Thread.Sleep(sleep);  // wait for powerup.  1 second works, but it is likely too much time...

            Thread.Sleep(1000);

            var start = PowerState.Uptime;

            // Wait till IRQ line goes low...  When the power is enabled, the cc3000 signals ready by IRQ going low
            while (!IsInitialized && irq.Read() == true && PowerState.Uptime.Subtract(start).Seconds < 4)
            {
                System.Threading.Thread.Sleep(20);
            }

            //Debug.Print("Interrupt state after start + 1s: " + irq.Read());

            if (irq.Read() == true && !IsInitialized) throw new Exception("CC3000 Power up failed.  The Interrupt (IRQ) failed to go low as it should have.  ");

            //Debug.Print("PowerUp enabled");
        }

        /// <summary>
        /// 
        /// </summary>
        internal void Start(byte mode = 0)
        {
            PowerUp();

            patchMode = mode;

            InitThread = new Thread(Init);
            InitThread.Start();
            //Debug.Print("IRQ state = " + irq.Read());

            InitWaiter.Reset();  // necessary with ARE?
            InitWaiter.WaitOne(15000, true);  // This makes the Constructor a blocking call while init completes

            patchMode = 0;

            if (!IsInitialized) throw new Exception("Failed to initialize the CC3000.  Startup aborted! ");

            WatchDogTimer = new Timer(new TimerCallback(PulseCheck), null, 15000, 2000);
        }

        void PollResponses()
        {
            IMessagePacket oldest = null;
            byte[] response = null;

            while (true)
            {
                //Debug.Print("Checking for responses");
                if (Responses.Any())
                {
                  //  Debug.Print("We have a response!");
                    oldest = Responses.Next;

                   //Debug.Print("We have a response = " + oldestKey);


                    if (oldest != null)
                    {
                        response = oldest.Content;

                      //  Debug.Print("Response length is " + response.Length);

                        Thread.Sleep(30);
                        Responses.Remove(oldest);
                        ResponseHandler(response); // Bad!!!  When an even fires, this thread is Consumed and never goes further!!  
                    }

                }
                else
                {
                    // wait 50ms for timeout and a set in the Read method...
                    //ResponseFlag.WaitOne(50, false);
                    //Thread.Sleep(50);
                    //Debug.Print("Waiting for set");
                    ResponseReadyFlag.WaitOne(100, false);
                    //Debug.Print("Waiting for set is over");
                }
            }
        }

        void PollInterrupt()
        {
            try
            {

                IMessagePacket oldest = null;

                while (true)
                {
                    //Debug.Print("Int Check");

                    try
                    {

                        if (!IsReading && !IsWriting && irq != null && !irq.Read() && IsInitialized)
                        {
                            Read();
                        }
                        else if (PendingCommands.Any())
                        {
                            oldest = PendingCommands.Next;

                            if (Write(oldest as PendingCommand)) PendingCommands.Remove(oldest);

                            if (oldest != null && oldest.WaitOpcode == (ushort)HciCommand.HCI_CMND_CLOSE_SOCKET)
                            {
                                // if we sent a Socket Close command, give a moment for a response
                                for (int i = 0; i <= 10; i++)
                                {
                                    Thread.Sleep(25);
                                    if (!irq.Read()) break;
                                }
                            }
                        }
                        else
                        {
                            //PendingCommands.Cleanup();
                            Thread.Sleep(15);
                        }

                        //Thread.Sleep(15);  // Added to allow other threads an opportunity to execute...
                    }
                    catch (Exception ex)
                    {
                        this.ResetAndReconnect();
                    }
                }
            }
            catch (Exception ex)
            {
                Debug.Print(ex.Message);
            }
        }


        void irq_OnInterrupt(object o)
        {

           // lock (irq)
           // {
            if (!IsReading && !IsWriting && irq != null && !irq.Read() && IsInitialized)
                {
                    //IsReading = true;
                    //Debug.Print("Interrupt -- READ!");
                    Read();
                }
           // }
            //else
            //{
                //Debug.Print("Interrupt!");
            //}

        }

        private void Init()
        {
            ///  The SPI transactions are mostly normal.  'Mostly', because:
            ///  the first one only after powerup requires a 50usec pause after the 4th byte before continuing.  seems like a workaround for device bug.  
            ///  I do this by two sequential WriteRead operations.  netmf is slow enough that you don't have to use any delay code, and you get about 84 us delay.  happily there doesn't appear to be a max delay.
            ///  the reads involve retrieving a length of the variably sized payload that follows, so you need to do those transactions in two phases also.
            ///  goofy first write (also this required to be a 'link start' command):

            try
            {

                Debug.Print(DateTime.Now.ToLocalTime() + " Sending - Simple Link Start");
                var response = this.Send(Command.Builder(HciCommand.HCI_CMND_SIMPLE_LINK_START, patchMode));

                if (response == null)
                    Debug.Print(DateTime.Now.ToLocalTime() + " CC3000 is not Responding to start-up commands.  ");
                
                //Debug.Print(DateTime.Now.ToLocalTime() + " Unknown firmware state.  cc3000 not responding with patches enabled.");

                if (!IsResetting && patchMode == 0) SetConnectionPolicy(false, false, true);

                if (patchMode == 0 && response != null) 
                {
                    // Mask out all non-required events from CC3000            
                    //Debug.Print("Setting Event Mask");
                    //this.Send(Command.Builder(HciCommand.HCI_CMND_EVENT_MASK, ((uint)(HciCommand.HCI_EVNT_DATA_UNSOL_FREE_BUFF | HciCommand.HCI_EVNT_WLAN_KEEPALIVE | HciCommand.HCI_EVNT_WLAN_ASYNC_PING_REPORT | HciCommand.HCI_EVNT_WLAN_UNSOL_INIT)).ToBytesRev()));
                    //this.Send(Command.Builder(HciCommand.HCI_CMND_EVENT_MASK, ((uint)(HciCommand.HCI_EVNT_DATA_UNSOL_FREE_BUFF | HciCommand.HCI_EVNT_WLAN_KEEPALIVE)).ToBytesRev()));

                    SetTimeouts();

                    if (FirmwareVersion[3] < 19) Debug.Print("Warning: This firmware version is old and will not work correctly with most UDP and TCP communications operations.  ");

                    Debug.Print("Requesting Buffer Size");
                    this.Send(Command.Builder(HciCommand.HCI_CMND_READ_BUFFER_SIZE));

                    ConfigureDHCP();

                    //SetLocalName(); 
                }
            }
            catch (Exception ex)
            {
                Debug.Print(ex.Message);
            }
            finally
            {
                InitWaiter.Set();  // Signal Init is complete!
            }
        }

        
        void PulseCheck(object o)
        {
            Debug.Print("*** Beat ***");
            Debug.Print("Queue: " + PendingCommands.Count + ", Unhandled Responses: " + Responses.Count);

            var silentTicks = PowerState.Uptime.Subtract(LastResponse).Ticks;

            if (!IsConnected || !IsInitialized || IsResetting || silentTicks < TimeSpan.TicksPerSecond * 2) return;

            Debug.Print("Watchdog Check");

            if (null == this.Send(Command.Builder(HciCommand.HCI_CMND_READ_BUFFER_SIZE), 400))
            {
                Debug.Print("Device appears dead, restarting...");
                this.WatchDogTimer.Change(Timeout.Infinite, Timeout.Infinite);  // disable watchdog

                // cc3000 is not responsive, reset the device
                this.ResetAndReconnect();
            }
        }

        internal byte[] IPAddress { get; set; }
        internal byte[] SubnetMask { get; set; }
        internal byte[] Gateway { get; set; }
        internal byte[] DnsServer { get; set; }        
        internal byte[] DhcpServer { get; set; }
        internal string ssid;

        private void ConfigureDHCP()
        {
            Debug.Print("Configuring DHCP");
            
            if (useDHCP || IPAddress == null)
                this.Send(Command.Builder(HciCommand.HCI_NETAPP_DHCP, 0, 0, 0, 0, 0));
            else
                this.Send(Command.Builder(HciCommand.HCI_NETAPP_DHCP, IPAddress, SubnetMask, Gateway, DnsServer));
        }

        private void RequestNetworkProperties()
        {
            Debug.Print("Requesting Network Connection details...");
            this.Send(Command.Builder(HciCommand.HCI_NETAPP_IPCONFIG), 7000);
        }

        private void ClearNetworkProperties()
        {
            IPAddress = null;
            SubnetMask = null;
            Gateway = null;
            this.DhcpServer = null;
            DnsServer = null;
            //MacAddress = new byte[6] { rawPayload[30], rawPayload[29], rawPayload[28], rawPayload[27], rawPayload[26], rawPayload[25] };
            //ssid = new string(Encoding.UTF8.GetChars(rawPayload, 31, 32));
        }

        private void PopulateNetworkProperties(byte[] rawPayload)
        {
            //ipconfig  This argument is a pointer to a
//!                         tNetappIpconfigRetArgs structure. This structure is
//!                         filled in with the network interface configuration.
//!                         tNetappIpconfigRetArgs:\n aucIP - ip address,
//!                         aucSubnetMask - mask, aucDefaultGateway - default
//!                         gateway address, aucDHCPServer - dhcp server address
//!                         aucDNSServer - dns server address, uaMacAddr - mac
//!                         address, uaSSID - connected AP ssid

            if (rawPayload == null || rawPayload.Length < 60)
            {
                Debug.Print("Warning: Invalid payload format for Network properties!");
                return;
            }

            try
            {
                IPAddress = new byte[4] { rawPayload[8], rawPayload[7], rawPayload[6], rawPayload[5] };
                SubnetMask = new byte[4] { rawPayload[12], rawPayload[11], rawPayload[10], rawPayload[9] };
                Gateway = new byte[4] { rawPayload[16], rawPayload[15], rawPayload[14], rawPayload[13] };
                this.DhcpServer = new byte[4] { rawPayload[20], rawPayload[19], rawPayload[18], rawPayload[17] };
                DnsServer = new byte[4] { rawPayload[24], rawPayload[23], rawPayload[22], rawPayload[21] };
                MacAddress = new byte[6] { rawPayload[30], rawPayload[29], rawPayload[28], rawPayload[27], rawPayload[26], rawPayload[25] };
                ssid = new string(Encoding.UTF8.GetChars(rawPayload, 31, 32));
            }
            catch
            {
                Debug.Print("Error parsing Network Properties Payload");
            }
        }

        internal byte[] FirmwareVersion
        {
            get
            {
                if (spVersion == null && IsInitialized)
                {
                    Debug.Print("Requesting SP Version");
                    var aResponse = this.Send(Command.Builder(HciCommand.HCI_CMND_READ_SP_VERSION));
                    if (aResponse == null) return null;
                    spVersion = Utility.ExtractRangeFromArray(aResponse, 5, 4);  // trim to just the relevant bytes
                }

                return spVersion;
            }
        }

        internal void ResetAndReconnect(bool silent = true)
        {
            SuppressConnectionEvent = silent;
            IsResetting = true;

            //SetConnectionPolicy(false, false, true);
            this.WatchDogTimer.Change(Timeout.Infinite, Timeout.Infinite);  // disable watchdog

            PendingCommands.Clear();
            Responses.Clear();
            BlockingHold.Set();  // release any blocking

            Thread.Sleep(2000);  // Wait for any blocking to finish?  

            this.Start();

            //SetConnectionPolicy(false, false, true);

            //this.PowerUp();
            //this.Send(Command.Builder(HciCommand.HCI_CMND_SIMPLE_LINK_START, 0));
            //ConfigureDHCP();

            //TODO: Add a timeout here: 
            while (!this.IsConnected || this.IPAddress == null) Thread.Sleep(50);

            foreach (var aListener in Listeners.Values)
            {
                if (aListener is TcpSocket) (aListener as TcpSocket).Reset();
                if (aListener is UdpSocket) (aListener as UdpSocket).Reset();
            }

            SuppressConnectionEvent = silent;
            IsResetting = false;
        }
        

        /// <summary>
        /// Read bytes from the Non-Volatile on-chip memory of the CC3000
        /// </summary>
        /// <param name="FileID">Choose a FileID from the Constants</param>
        /// <param name="start">Start byte position with the file</param>
        /// <param name="length">Length of bytes to retrieve</param>
        /// <returns></returns>
        internal byte[] NonVolatileMemoryRead(int FileID, int start, int length)
        {
            // the reason I am using locking and blocking here is because if we allow multiple simultaneous calls, 
            // we won't be able to tell which response went with which query.  
            // Although this could still happen if we ever have a response that exceeds the timeout, but oh well...
            
            lock (nvWaiter)
            {
                nvBuffer = null;

                this.Send(Command.Builder(HciCommand.HCI_CMND_NVMEM_READ, FileID, length, start));

                nvWaiter.WaitOne(5000, true);
                return nvBuffer;
            }
        }


        internal void CreateEntry(int FileID, int length)
        {
            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_NVMEM_CREATE_ENTRY, FileID, length));

            if (response == null || response.Length < 8 || response.ToInt(4) != 0)
                throw new Exception("Failed to create File Entry");
        }

        /// <summary>
        /// Write a byte array to the Non-Volatile on-chip memory of the CC3000
        /// </summary>
        /// <param name="FileID">Choose a FileID from the Constants</param>
        /// <param name="start">Start byte position with the file</param>
        /// <param name="data">Data array to write</param>
        /// <returns></returns>
        internal int NonVolatileMemoryWrite(int FileID, int start, byte[] data)
        {
            if (!WriteEnabled)
            {
                Debug.Print("Write is Disabled: tried to write " + data.Length + " bytes (" + Utility.ExtractRangeFromArray(data, 0, System.Math.Min(data.Length, 3)).ToHexString() + "..." + Utility.ExtractRangeFromArray(data, System.Math.Max(0, data.Length - 3), System.Math.Min(data.Length, 3)).ToHexString() + ") to file: " + FileID);
                return 0;
            }


            var response = this.Send(Command.DataBuilder(HciCommand.HCI_CMND_NVMEM_WRITE, data, null, FileID, 12, data.Length, start), (ushort)HciCommand.HCI_EVNT_NVMEM_WRITE, 5000);

            if (response == null || response.Length < 8) return -1;

            return (int)response.ToInt(4);
            
        }

        /// <summary>
        /// Takes the whole file as a byte array and writes it into the non-volatile memory in bite sized chunks
        /// </summary>
        /// <param name="FileID"></param>
        /// <param name="data"></param>
        /// <returns></returns>
        internal int NvMemWriteFile(int FileID, byte[] data, int offsetStart = 0)
        {
            int status = 0;
            int offset = offsetStart;
            int maxChunkSize = 24; // 150;  //  should be 20?  seems small, so I am using a value from comments suggesting 150
            int currentChunkSize = maxChunkSize;
            int numberOfBytesToSend = data.Length;

            Debug.Print("Writing File " + FileID);

            while ((status == 0) && (numberOfBytesToSend > 0))
	        {
                currentChunkSize = System.Math.Min(maxChunkSize, numberOfBytesToSend);
                if (FileID == 16 && offset == 0) currentChunkSize = 2;  // For the FAT, write the first 2 bytes seperately?
                status = NonVolatileMemoryWrite(FileID, offset, Utility.ExtractRangeFromArray(data, offset, currentChunkSize));
                offset += currentChunkSize;
                numberOfBytesToSend -= currentChunkSize;
	        }

	        return status;   // 0 is success.  
        }



        /// <summary>
        /// This is an synchronous command to read the Mac address and populate the property with the response 
        /// </summary>
        /// <returns></returns>
        private void ReadMacAddress()
        {
            // Note: PopulateNetworkProperties does this, so there should be no need to use this... 
            // although it's a great+working example of reading the NV memory

            var result = NonVolatileMemoryRead(Constants.NVMEM_MAC_FILEID, 0, 6);

            if (result != null && result.Length >= 6) 
                MacAddress = Utility.ExtractRangeFromArray(result, result.Length - 6, 6);  // the Mac is in the last 6 bytes
        }

        /// <summary>
        /// Allows the CC3000 resolve to a local name using mDNS
        /// </summary>
        /// <param name="name"></param>
        internal void SetLocalName()
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_CMND_MDNS_ADVERTISE, 1, 8, Name.Length, Name)); 

            //TODO: Support LLMNR with a socket in the future?
        }

        /// <summary>
        /// Send a Ping request (ICMP) to network host
        /// Note: Calling this function while a previous Ping Requests are in progress will stop the previous ping request.
        /// </summary>
        /// <param name="address">destination IP address (4-byte array)</param>
        /// <param name="attempts">number of echo requests to send</param>
        /// <param name="size">send buffer size which may be up to 1400 bytes</param>
        /// <param name="timeout">Time to wait for a response, in milliseconds.</param>
        internal void Ping(byte[] address, uint attempts, uint size, uint timeout)
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_PING_SEND, address, attempts, size, timeout)); 
        }

        internal void GetPingStatus()
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_PING_REPORT)); 
        }

        internal void PingStop()
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_PING_STOP)); 
        }

        internal void GetIpConfig()
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_IPCONFIG)); 
        }

        /// <summary>
        /// Set new timeout values
        /// Note: If a parameter set to non zero value which is less than 20s, it will be set automatically to 20s
        /// </summary>
        /// <param name="dhcpTimeout">DHCP lease time request, also impacts the DHCP renew timeout.  Set to 0 for Infinity.  </param>
        /// <param name="arpTimeout">ARP refresh timeout, if ARP entry is not updated by incoming packet, the ARP entry will be  deleted by the end of the timeout.  Set to 0 for Infinity. </param>
        /// <param name="keepAlive">Keepalive event sent by the end of keepalive timeout.  Set to 0 for Infinity. </param>
        /// <param name="inactivity">Socket inactivity timeout, socket timeout is refreshed by incoming or outgoing packet, by the end of the socket timeout the socket will be closed.  Set to 0 for Infinity. </param>
        internal void SetTimeouts(uint dhcpTimeout = 14400, uint arpTimeout = 3600, uint keepAlive = 10, uint inactivity = 0)
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_SET_TIMERS, dhcpTimeout, arpTimeout, keepAlive, inactivity)); 
        }

        internal void FlushArp()
        {
            //TODO: This needs to be tested

            this.Send(Command.Builder(HciCommand.HCI_NETAPP_ARP_FLUSH)); 
        }


        /// <summary>
        /// Connect to the desired Wifi AccessPoint
        /// </summary>
        /// <param name="ap"></param>
        /// <returns></returns>
        public bool Connect(AccessPoint ap)
        {
            if (IsInitialized)
            {
                Debug.Print("Resetting Connection before setting Policy. ");
                Disconnect();
            }

            SetConnectionPolicy(false, false, false);

            // This sleep was recommended on the TI forums... no clue if it is necessary :)
            //Debug.Print("Sleeping for 8s");
            //Thread.Sleep(8000);


            Debug.Print("Attempting Connection to " + ap.Ssid);
            this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_CONNECT, 0x0000001c, ap.Ssid.Length, (int)ap.Security, (ap.Ssid.Length + 16), ap.Key.Length, (short)0, new byte[6], ap.Ssid, ap.Key), 12000);

            return true;  // response not implemented
        }

        /// <summary>
        /// Disconnect from Wifi
        /// </summary>
        /// <returns></returns>
        public void Disconnect()
        {
            this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_DISCONNECT));

            ClearNetworkProperties();

            //MacAddress = null;
            //IPAddress = null;

            //return true;  // response not implemented
        }

        
        /// <summary>
        /// Turns off the cc3000 device (waits up to 2 seconds for the shutdown) then returns
        /// </summary>
        /// <returns>True for success, false for failure...</returns>
        public bool Shutdown()
        {
            //Debug.Print("Interrupt state before shutdown: " + irq.Read());

            Debug.Print("Shutting down CC3000...");

            if (irq.Read() && pwren.Read())
            {
                pwren.Write(false);
                System.Threading.Thread.Sleep(100); // wait for shutdown?
            }

            Responses.Clear();
            PendingCommands.Clear();

            cs.Write(true);

            pwren.Write(false);  // signal the power down
            IsInitialized = false;

            //Debug.Print("Interrupt state after pwr disabled: " + irq.Read() + ", " + PowerState.Uptime.ToString());

            //Debug.Print("Interrupt state after pwr disabled: " + irq.Read() + ", " + PowerState.Uptime.ToString());

            var start = PowerState.Uptime;

            
            // Wait till IRQ line goes high...
            do
                System.Threading.Thread.Sleep(20);
            while (irq.Read() == false && PowerState.Uptime.Subtract(start).Seconds < 4);


            Thread.Sleep(300);

            if (IsResetting) Thread.Sleep(1500);  // Wait for Caps to discharge?

            //Debug.Print("Interrupt state: " + irq.Read() + ", " + PowerState.Uptime.ToString());

            //Close spi?

            return irq.Read();  
        }

        /// <summary>
        /// Set Connection Policy
        /// </summary>
        /// <param name="ConnectToOpenAccessPoints">enable(true), disable(false) to connect to any available AP</param>
        /// <param name="UseFastConnect">enable(true), disable(false). if enabled, tries to connect to the last connected AP. </param>
        /// <param name="UseProfiles">enable(true), disable(false) auto connect after reset and periodically reconnect if needed.</param>
        /// <returns>true for success, false for failure</returns>
        public void SetConnectionPolicy(bool ConnectToOpenAccessPoints, bool UseFastConnect, bool UseProfiles)
        {
            
            ///     When auto is enabled, the device tries to connect according
            ///              the following policy:
            ///              1) If fast connect is enabled and last connection is valid,
            ///                 the device will try to connect to it without the scanning
            ///                 procedure (fast). The last connection will be marked as
            ///                 invalid, due to adding/removing profile.
            ///              2) If profile exists, the device will try to connect it
            ///                 (Up to seven profiles).
            ///              3) If fast and profiles are not found, and open mode is
            ///                 enabled, the device will try to connect to any AP.
            ///              * Note that the policy settings are stored in the CC3000 NVMEM.
            ///

            Debug.Print("Setting Connection Policy.");

            // Note: using the "ToInt()" method here because each argument must be 4 bytes each
            this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_SET_CONNECTION_POLICY, ConnectToOpenAccessPoints.ToInt(), UseFastConnect.ToInt(), UseProfiles.ToInt()));
        }


        public void DeleteProfile(byte profileID)
        {
            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_DEL_PROFILE, (int)profileID), 3000);

            if (response == null || response.Length < 8 || response.ToInt(4) != 0)
                throw new Exception("Failed to Delete Profile.");
        }

        public void DeleteAllProfiles()
        {
            DeleteProfile(255);
        }

        /// <summary>
        /// Start Listening for the smart config response
        /// </summary>
        /// <param name="timeout">time to wait, in seconds</param>
        /// <param name="useEncryption">Use Encryption?</param>
        public void StartSmartConfig(int timeout = 10, string encryptionKey = null)
        {
            Debug.Print("Waiting for the Magic Access Point packet to arrive...");

            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_START, encryptionKey == null ? 0 : 1), (ushort)HciCommand.HCI_EVNT_WLAN_ASYNC_SIMPLE_CONFIG_DONE, (uint)timeout * 1000);

            if (response == null || response.Length < 5 || response[4] != 0)
            {
                StopSmartConfig();
                throw new Exception("Failed to Receive a SmartConfig Signal!");
            }
        }

        /// <summary>
        /// Stop Listening for the smart config response
        /// </summary>
        internal void StopSmartConfig()
        {
            this.SendAsync(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_STOP));
        }

        public void SetSmartConfigPrefix()
        {
            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_SET_PREFIX, "TTT"));

            if (response == null || response.Length < 8 || response.ToInt(4) != 0)
                throw new Exception("Failed to set SmartConfig Prefix.");
        }

        public void ProcessSmartConfig(string encryptionKey = null)
        {
            var accessPointInfo = NonVolatileMemoryRead(Constants.NVMEM_SHARED_MEM_FILEID, 0, 67);

            if (accessPointInfo == null || accessPointInfo.Length < 67) throw new Exception("Failed to Read Access Point Info");

            var ssidLength = accessPointInfo[0];
            var passwordLength = accessPointInfo[ssidLength + 1];
            var security = (WifiSecurityType)accessPointInfo[ssidLength + 2];
            var ssidBytes = Utility.ExtractRangeFromArray(accessPointInfo, 1, ssidLength);
            var ssid = new string(Encoding.UTF8.GetChars(ssidBytes));
            var passwordBytes = Utility.ExtractRangeFromArray(accessPointInfo, ssidLength + 3, passwordLength);
            string password = string.Empty;

            if (encryptionKey != null)
            {
                // Using AES encryption!  
                var key = this.NonVolatileMemoryRead(Constants.NVMEM_AES128_KEY_FILEID, 0, 16);

                //TODO: Decrypt Password here from passwordBytes and private shared key
                var decryptedBytes = Security.AesDecrypt(passwordBytes, key);

                password = new string(Encoding.UTF8.GetChars(decryptedBytes));
            }
            else
            {
                // password is already unscrambled
                password = new string(Encoding.UTF8.GetChars(passwordBytes));
            }

            switch (security)
            {
                case WifiSecurityType.WEP:
                    AddProfile(ssid, security, password);
                    break;
                case WifiSecurityType.WPA:
                case WifiSecurityType.WPA2:
                    AddProfile(ssid, security, password);
                    break;
                default:
                    AddProfile(ssid);
                    break;
            }
        }

        public void AddProfile(string ssid, WifiSecurityType security = WifiSecurityType.Open, string password = null, int priority = 1, byte[] bssid = null)
        {
            Debug.Print("Adding Profile for SSID " + ssid + ", pass: " + password);

            byte[] response = null;

            if (bssid == null || bssid.Length != 6) bssid = new byte[6];
            
            switch (security)
            {
                case WifiSecurityType.WEP:
                    response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_ADD_PROFILE, (uint)security, 32, ssid.Length, (short)0, bssid, priority, ssid.Length + 12, password.Length, 0, ssid, password, password, password, password), 4000);
                    break;
                case WifiSecurityType.WPA:
                case WifiSecurityType.WPA2: 
                    response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_ADD_PROFILE, (uint)security, 40, ssid.Length, (short)0, bssid, priority, 24, 30, 2, ssid.Length + 8, password.Length, ssid, password), 4000);
                    break;
                default:
                    response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_ADD_PROFILE, (uint)security, 20, ssid.Length, (short)0, bssid, priority, ssid), 4000);
                    break;
            }

            // Frequently, add_profile does not return a response, but it seems to always work, so I have to assume that this worked...
            
            //if (response == null || response.Length < 8 || response.ToInt(4) != 0)
            //    throw new Exception("Failed to add profile for ssid " + ssid);
        }

        /// <summary>
        /// Scan for available Wifi networks and return them as an array
        /// This is a Synchronous call and will block execution
        /// </summary>
        public AccessPointResult[] NetworkScan(int Max = 5, uint minDwell = 20, uint maxDwell = 30)
        {
            // http://e2e.ti.com/support/low_power_rf/f/851/p/186022/671001.aspx
            //The wlan_ioctl_get_scan_results() return the populated table inside the CC3000 of the surrounding APs.

            //The CC3000 populate this table according to generated scan policies.

            //Few terminologies explanation: 

            //Dwell time - The amount of time the CC3000 sends probe request and listen to probe response/beacons per channel. It's important that this time will be long enough for the STA to hear all the APs on that channel. Advise to use defaults
            //Channels mask - There are up to 13 channel, However in some countries regulatory domains prevent you from usage of a channel
            //Scan cycle - Going for all supported channels and allocating the information for that. Calculated time: num of channel*dwell time
            //Interval - The amount of time between scan-cycle. 
            //There are 2 different policies:

            //While the CC3000 is not connected to any AP. 
            //In that case the CC3000 issues a "scan cycle" wait "interval time[cycle number]" The last entry in the list will be repeated until a connection. 
            //The purpose of that is to try to connect very fast to an AP, and if there are non, increase the interval to save TX power.
            //While the CC3000 is connected. In that case only a single interval is being used. Not sure exactly which one. (Igor?) 
            //- Elad            

            
            var timeoutArray = new UInt32[Constants.SL_SET_SCAN_PARAMS_INTERVAL_LIST_SIZE];
            for (int i = 0; i < timeoutArray.Length; i++) timeoutArray[i] = 2000;  // 2000ms

            this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_SET_SCANPARAM, 36, 1, minDwell, maxDwell, 2, 2047, -100, 0, 205, timeoutArray));

            var results = new Hashtable();

            for (int i = 0; i < Max; i++)
            {
                var aResult = GetScanResult();
                if (aResult != null && !results.Contains(aResult.Ssid)) results.Add(aResult.Ssid, aResult);
            }

            var resultArray = new AccessPointResult[results.Count];

            //for (int i = 0; i < resultArray.Length; i++)
            //    resultArray[i] = results[`i] as AccessPointResult;

            int j = 0;

            foreach (var aNetwork in results.Values)
                resultArray[j++] = aNetwork as AccessPointResult;

            return resultArray;
        }

        private AccessPointResult GetScanResult()
        {


            //long wlan_ioctl_get_scan_results(unsigned long ulScanTimeout,
            //                                   unsigned char *ucResults)
            //#endif /* __ENABLE_MULTITHREADED_SUPPORT__ */
            //{
            //    unsigned char *ptr;
            //    unsigned char *args;

            //    ptr = tSLInformation.pucTxCommandBuffer;
            //    args = (ptr + HEADERS_SIZE_CMD);

            //    // Fill in temporary command buffer
            //    args = UINT32_TO_STREAM(args, ulScanTimeout);

            //    // Initiate a HCI command
            //    hci_command_send(HCI_CMND_WLAN_IOCTL_GET_SCAN_RESULTS,
            //        ptr, WLAN_GET_SCAN_RESULTS_PARAMS_LEN);

            //    // Wait for command complete event
            //    SimpleLinkWaitEvent(HCI_CMND_WLAN_IOCTL_GET_SCAN_RESULTS, ucResults);

            //    return(0);

            //!  @brief    Gets entry from scan result table.
            //!            The scan results are returned one by one, and each entry
            //!            represents a single AP found in the area. The following is a
            //!            format of the scan result:
            //!          - 4 Bytes: number of networks found
            //!          - 4 Bytes: The status of the scan: 0 - aged results,
            //!                     1 - results valid, 2 - no results
            //!          - 42 bytes: Result entry, where the bytes are arranged as  follows:
            //!
            //!                         - 1 bit isValid - is result valid or not
            //!                         - 7 bits rssi - RSSI value;
            //!                 - 2 bits: securityMode - security mode of the AP:
            //!                           0 - Open, 1 - WEP, 2 WPA, 3 WPA2
            //!                         - 6 bits: SSID name length
            //!                         - 2 bytes: the time at which the entry has entered into
            //!                            scans result table
            //!                         - 32 bytes: SSID name
            //!                 - 6 bytes:  BSSID

            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_WLAN_IOCTL_GET_SCAN_RESULTS, 1));

            var status = response.ToInt(9);  // 0 = aged, 1 = valid, 2 = no result

            var networksRemaining = response.ToInt(5);  // not used

            if (status != 1) return null;

            try
            {
                var isValid = (response[13] & 0x01) == 1;
                if (!isValid) return null;

                var security = (WifiSecurityType)(response[14] & 0x03);
                var ssidSize = (response[14] & 0xFC) >> 2;

                var hostspotFindTime = response.ToShort(15);  // not used
                var ssidName = Utility.ExtractRangeFromArray(response, 17, ssidSize);

                var rssi = (response[13] & 0xFE) >> 1;
                var bssid = Utility.ExtractRangeFromArray(response, 49, 6);

                return new AccessPointResult() { Ssid = new string(Encoding.UTF8.GetChars(ssidName)), Security = security, SignalStrength = (sbyte)rssi, MacAddress = bssid };
            }
            catch
            {
                // parsing error!
                return null;
            }
        }

        /// <summary>
        /// Get host IP by name. Obtain the IP Address of machine on network by its name
        /// </summary>
        /// <param name="hostname"></param>
        /// <returns>IP Address</returns>
        internal byte[] LookupAddress(string hostname)
        {
            //TODO: Verify that this method works

            hostname = hostname.Trim().ToLower();

            var response = this.Send(Command.Builder(HciCommand.HCI_CMND_GETHOSTNAME, 8, hostname.Length, hostname));

            if (response == null || response.Length < 13) return null;

            return new byte[] { response[12], response[11], response[10], response[9] } ;
        }


        /// <summary>
        /// Read an incoming message from the CC3000
        /// </summary>
        private void Read()
        {
           // IsReading = true;
          //  if (packetCount > 100) return;

            //reading is a little interesting in that you must retrieve the length, and then complete the reading:
           // Debug.Print("ReMem1: " + MemDelta(true));

            var responseHeader = new byte[10];

           // Debug.Print("ReMem2: " + MemDelta(true));
            
            byte[] pktReadPayload = null;

            //Debug.Print("ReMem3: " + MemDelta(true));

            //Debug.Print("Read Waiting");

              //  Debug.Print("ReMem4: " + MemDelta(true));

            //lock (spiLock)
            //{
                //wait for the signal that its ready
                //TODO: support a timeout
                //while (irq.Read()) { }
                if (irq.Read()) return;

                //assert cs and (redundantly) wait for irq
                cs.Write(false);

                var pktReadHeader = new byte[10] { 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF };

                //Debug.Print("ReMem5: " + MemDelta(true));


                //write header and read the length bytes
                spi.WriteRead(pktReadHeader, 0, 10, responseHeader, 0, responseHeader.Length, 0);



                //var opcode = (short)((responseHeader[7] << 8) | responseHeader[6]);
                var payloadSize = (short)((responseHeader[3] << 8) | responseHeader[4]);
                //    var oneByteArgSize = responseHeader[5] == (byte)HciCommandType.HCI_TYPE_EVNT;

                //make length into a number; header is big endian.  other things are little.
                //var argsLength = (responseHeader[9] << 8) | responseHeader[8];
                //var argsLength = oneByteArgSize ? responseHeader[8] : (responseHeader[9] << 8) | responseHeader[8];

                //Debug.Print("ReMem6: " + MemDelta(true));

                //Debug.Print("Size of pktReadPayload: " + payloadSize);

                if (payloadSize > 2048)
                {
                    // Throw an exception here? 
                    Debug.Print("Large Payload.  Likely a bad read");

                }
                else if (payloadSize > 0)
                {

                    pktReadPayload = new byte[payloadSize];
                    pktReadPayload[0] = responseHeader[5];  // grab the first 5 bytes of payload
                    pktReadPayload[1] = responseHeader[6];
                    pktReadPayload[2] = responseHeader[7];
                    pktReadPayload[3] = responseHeader[8];
                    pktReadPayload[4] = responseHeader[9];

                    //Debug.Print("ReMem7: " + MemDelta(true));


                    if (payloadSize > 5)
                    {
                        //read payload

                        //TODO: sanity check on length

                        spi.WriteRead(pktReadHeader, 0, 3, pktReadPayload, 5, payloadSize - 5, 0);

                        //if (oneByteArgSize)
                        //    spi.WriteRead(pktReadHeader, 0, 3, pktReadPayload, 1, argsLength-1, 0);
                        //else
                        //    spi.WriteRead(pktReadHeader, 0, 3, pktReadPayload, 0, argsLength, 0);

                    }
                }
                //Debug.Print("ReMem8: " + MemDelta(true));

                // irq.ClearInterrupt();


                //negate cs
                cs.Write(true);

                //IsReading = false;
           // }

            //packetCount++;

                if (pktReadPayload != null && pktReadPayload.Length >= 4)
                {
                    // write opcode and raw payload to Response collection

                    var opcode = (ushort)((pktReadPayload[2] << 8) | pktReadPayload[1]);

                    Debug.Print(DateTime.Now.ToLocalTime() + " New Response Received (" + opcode.ToHexString() + ") - " + ((pktReadPayload.Length > 20) ? Utility.ExtractRangeFromArray(pktReadPayload, 0, 20).ToHexString() + " ..." : pktReadPayload.ToHexString()));

                    // Special socket specific opcode substitution!  
                    if (opcode == (ushort)HciCommand.HCI_CMND_ACCEPT)
                    {
                        ushort newOpcode = (ushort)(0x9000 + pktReadPayload.ToInt(5));
                        Debug.Print("Adding special Opcode: " + newOpcode.ToHexString());
                        Responses.Add(new ReceivedResponse() { WaitOpcode = newOpcode, Content = pktReadPayload} );
                    }
                    else if (opcode == (ushort)HciCommand.HCI_CMND_RECV)
                    {
                        ushort newOpcode = (ushort)(0x9010 + pktReadPayload.ToInt(5));
                        Debug.Print("Adding special Opcode: " + newOpcode.ToHexString());
                        Responses.Add(new ReceivedResponse() { WaitOpcode = newOpcode, Content = pktReadPayload });
                        //Responses.Add(newOpcode, pktReadPayload);
                    }
                    else if (opcode == (ushort)HciCommand.HCI_CMND_RECVFROM)
                    {
                        ushort newOpcode = (ushort)(0x9020 + pktReadPayload.ToInt(5));
                        Debug.Print("Adding special Opcode: " + newOpcode.ToHexString());
                        //Debug.Print(pktReadPayload.ToHexString());
                        Responses.Add(new ReceivedResponse() { WaitOpcode = newOpcode, Content = pktReadPayload });

                        //Responses.Add(newOpcode, pktReadPayload);
                    }
                    else if (opcode == (ushort)HciCommand.HCI_EVNT_BSD_TCP_CLOSE_WAIT)
                    {
                        sockets[pktReadPayload.ToInt(5)].IsClosing = true;
                        Responses.Add(new ReceivedResponse() { WaitOpcode = (ushort)((pktReadPayload[2] << 8) | pktReadPayload[1]), Content = pktReadPayload });

                        //Responses.Add((ushort)((pktReadPayload[2] << 8) | pktReadPayload[1]), pktReadPayload);
                    }
                    else
                    {
                        if (opcode == (ushort)HciCommand.HCI_DATA_RECV || opcode == (ushort)HciCommand.HCI_DATA_RECVFROM)
                            Debug.Print("GotData - " + pktReadPayload.ToHexString());

                        Responses.Add(new ReceivedResponse() { WaitOpcode = (ushort)((pktReadPayload[2] << 8) | pktReadPayload[1]), Content = pktReadPayload });
                        //Responses.Add((ushort)((pktReadPayload[2] << 8) | pktReadPayload[1]), pktReadPayload);
                    }

                    //else if (opcode == (ushort)HciCommand.HCI_DATA_RECV)
                    //{
                    //    Debug.Print("GotData - " + pktReadPayload.ToHexString());
                    //}
                    //else if (opcode == (ushort)HciCommand.HCI_DATA_RECVFROM)
                    //{
                    //    Debug.Print("GotData - " + pktReadPayload.ToHexString());
                    //}  

                    BlockingHold.Set();
                    ResponseReadyFlag.Set();
                }
                else
                {
                    Debug.Print("Small packet?");
                }





            //ResponseHandler()

            //TODO: spawn a thread here for the handler?
            //if (pktReadPayload != null) ResponseHandler(pktReadPayload);

            //Debug.Print("ReMem9: " + MemDelta(true));

        }

        /// <summary>
        /// Asynchronous version of the Send command method.  This method will return immediately and will not wait for a response.  
        /// </summary>
        /// <param name="theCommand"></param>
        internal void SendAsync(byte[] theCommand)
        {
            this.Send(theCommand, 0);
        }

        /// <summary>
        /// Asynchronous version of the Send command method.  This method will return immediately and will not wait for a response.  
        /// </summary>
        /// <param name="theCommand"></param>
        internal byte[] Send(byte[] theCommand, uint timeout = 3000)
        {
            return this.Send(theCommand, Command.GetOpCode(theCommand), timeout);
        }

        /// <summary>
        /// Send a command to the CC3000.  Note, this is a blocking (aka. synchronous) call with a 2 second timeout for the response, unless otherwise specified
        /// </summary>
        /// <param name="theCommand">The command object</param>
        /// <param name="timeout">How long to wait for a response, in milliseconds</param>
        internal byte[] Send(byte[] theCommand, ushort waitOpcode, uint timeout)
        {
            PendingCommands.Add(new PendingCommand() { Content = theCommand, WaitOpcode = waitOpcode} );

            var start = PowerState.Uptime;

            if (waitOpcode == 0x9000) Debug.Print("Accept Send @ " + start.Seconds + "s");


            //TODO: This timeout should exclude other commands in the queue before the executed command
            // If the timeout is not 0, wait here until we have a response or we timeout
            while (timeout > 0 && Responses.Any(waitOpcode) == false && PowerState.Uptime.Subtract(start) < new TimeSpan(0, 0, 0, 0, (int)timeout))
                BlockingHold.WaitOne(30, false);

            if (waitOpcode == 0x9000) Debug.Print("Accept Elapse @ " + start.Seconds + "s - Response? " + Responses.Any(waitOpcode).ToString());

                //Thread.Sleep(15);

            if (Responses.Any(waitOpcode) && timeout > 0) 
            {
                byte[] theResponse = Responses.NextWhere(waitOpcode).Content;
                //Responses.Remove(waitOpcode);

                // we have a response!
                return theResponse;
            }

            // we timed out, so return null...
            return null;
        }


        private bool Write(PendingCommand pendingCommand)
        {

            Debug.Print(DateTime.Now.ToLocalTime() + " Sending (" + pendingCommand.WaitOpcode.ToHexString() + ") - " + pendingCommand.Content.ToHexString());

            if (irq == null) return false;
            if (pendingCommand == null) return true;  // we have successfully written a null command... because there is nothing to write for a null command :)

           // lock (spiLock)
           // {
                var start = PowerState.Uptime;

                if (!IsInitialized)
                {
                    cs.Write(false);

                    System.Threading.Thread.Sleep(1);  // wait at least 50us according to TI docs

                    while (irq.Read() && PowerState.Uptime.Subtract(start).Seconds < 5) { Thread.Sleep(50); }
                    if (irq.Read())
                    {
                        // we timed out because a read is not pending, so we have to get out of the lock to allow the Read to happen...
                        cs.Write(true);
                        //IsWriting = false;

                        Shutdown();

                        throw new Exception("Initialization Error.  CC3000 failed to respond to the power up with the usual Active High on the Interrupt pin.  Check all connections.");
                    }
                }
                //else
                //{
                //    //while (!irq.Read() && PowerState.Uptime.Subtract(start).Seconds < 5) { }
                //    if (IsInitialized && irq.Read())
                //        goto tryAgain;  // we timed out because a read is pending, so we have to get out of the lock to allow the Read to happen...
                //}

                //Debug.Print("Send " + opcode.ToHexString() + " isWriting");

                if (!IsInitialized || irq.Read())
                {
                    //if (waitOpcode != 0 && timeout > 0)
                    //{
                    //    if (WaiterOpCodes.Contains(waitOpcode))
                    //        WaiterOpCodes[waitOpcode] = PowerState.Uptime;
                    //    else
                    //        WaiterOpCodes.Add(waitOpcode, PowerState.Uptime);
                    //}

                    //IsWriting = true;

                    //assert cs and (redundantly) wait for irq
                    cs.Write(false);

                    if (!IsInitialized) System.Threading.Thread.Sleep(1);  // sleep at least 50 MicroSeconds here

                    //wait for the signal that its ready
                    //TODO: support a timeout
                    while (irq.Read()) { Thread.Sleep(1); }

                    //Debug.Print(Microsoft.SPOT.Hardware.Utility.GetMachineTime().Seconds + " - Sending Command: " + ((short)theCommand.OpCode).ToHexString());

                    int chunkSize = IsInitialized ? 512 : 4;  // the initial chunksize must be 4 according to the docs
                    int transmitSize = 0;
                    int bytesTrasmitted = 0;

                    //byte[] commandToTransmit = theCommand;
                    //byte[] response = new byte[chunkSize];

                    while (bytesTrasmitted < pendingCommand.Content.Length)
                    {
                        transmitSize = System.Math.Min(chunkSize, pendingCommand.Content.Length - bytesTrasmitted);
                        //var msg = Utility.ExtractRangeFromArray(commandToTransmit, bytesTrasmitted, transmitSize);
                        //response = new byte[transmitSize];
                        //write header and read the length bytes
                        //spi.WriteRead(msg, 0, transmitSize, response, 0, transmitSize, 0);
                        //spi.Write(msg);

                        if (IsInitialized)
                        {
                            spi.WriteRead(pendingCommand.Content, bytesTrasmitted, transmitSize, null, 0, 0, 0);
                        }
                        else
                        {
                            var response = new byte[transmitSize];
                            spi.WriteRead(pendingCommand.Content, bytesTrasmitted, transmitSize, response, 0, transmitSize, 0);

                            if (!response.BytesEqual(ExpectedSimpleLinkResponse))
                            {
                                cs.Write(true);
                                //IsWriting = false;

                                Shutdown();

                                throw new Exception("SPI response to Simple Link Start was invalid, check SPI connection to CC3000. Response was " + response.ToHexString());
                            }
                        }

                        IsInitialized = true;
                        bytesTrasmitted += transmitSize;
                        chunkSize = 512;
                    }

                    //if (timeout > 0 && waitOpcode != 0)
                    //{
                    //    if (responses.Contains(waitOpcode)) responses.Remove(waitOpcode);
                    //    WaiterOpCode = waitOpcode;
                    //}

                    // Wait for response

                    cs.Write(true);
                    //IsWriting = false;

                    //negate cs

                    return true;
                }

                return false;
            //}
        }


        /// <summary>
        /// Do the appropriate thing with the response from the CC3000
        /// </summary>
        /// <param name="aResponse"></param>
        internal void ResponseHandler(byte[] aResponse)
        {
            int socketID = -1;

            var opcode = (ushort)((aResponse[2] << 8) | aResponse[1]);

            Debug.Print("Handling Response for OpCode: " + opcode.ToHexString() + ", Payload: " + aResponse.ToHexString());
            
            
            //TODO: this should probably be socket specific...  if socket 1 is hosed, but socket 2 is still getting responses, 
            //      Socket 1 will never hit it's socket timeout!!!!  Right?
            LastResponse = PowerState.Uptime;

            // do opcode specific stuff

            switch ((HciCommand)opcode)
            {
                case HciCommand.HCI_CMND_SIMPLE_LINK_START: Debug.Print("Response Received -- Simple Link Start Confirmed"); break;
                case HciCommand.HCI_CMND_EVENT_MASK: Debug.Print("Response Received -- Event Mask confirmed"); break;
                case HciCommand.HCI_CMND_READ_BUFFER_SIZE: Debug.Print("Response Received -- Buffer Available: " + aResponse[5].ToString() + ", Size: " +  new byte[2] { aResponse[6], aResponse[7] }.ToShort()); break;
                case HciCommand.HCI_CMND_READ_SP_VERSION:
                    aResponse = Utility.ExtractRangeFromArray(aResponse, 5, 4);  // trim to just the relevant bytes
                    Debug.Print("Response Received -- SP Version is " + aResponse.ToAddress() ); break;
                case HciCommand.HCI_EVNT_WLAN_UNSOL_CONNECT: 
                    Debug.Print("WLAN Connection Established!  Waiting for IP assignment from router... " + aResponse.ToHexString(5));
                    IsConnected = true;
                    break;
                case HciCommand.HCI_CMND_WLAN_IOCTL_GET_SCAN_RESULTS:
                case HciCommand.HCI_CMND_WLAN_IOCTL_SET_SCANPARAM:
                    // Nothing to do here
                    break;
                case HciCommand.HCI_EVNT_WLAN_UNSOL_DHCP:
                    if (aResponse[8] == 0)
                    {
                        ClearNetworkProperties();
                        break;
                    }

                    // Parse this response and populate properties, like ip address and gateway!  
                    Debug.Print("IP Address Acquired => " + new byte[4] { aResponse[8], aResponse[7], aResponse[6], aResponse[5] }.ToAddress()); //  aResponse.Payload.ToHexString()); 
                    //SetLocalName();  // establish local name
                    
                    this.RequestNetworkProperties();  // This will fill in the detailed properties
                    break;
                case HciCommand.HCI_EVNT_NVMEM_WRITE:
                    // nothing to do
                    break;
                case HciCommand.HCI_DATA_NVMEM:
                    if (aResponse != null && aResponse.Length > 29)
                        this.nvBuffer = Utility.ExtractRangeFromArray(aResponse, 29, aResponse.Length - 29);
                    else
                        this.nvBuffer = null;

                    nvWaiter.Set();  // This will release the waiting nvMem read
                    break;
                case HciCommand.HCI_EVNT_PATCHES_REQ:
                    switch ((PatchType)(aResponse[5] | aResponse[6] << 8))
                    {
                        case PatchType.HCI_EVENT_PATCHES_BOOTLOAD_REQ:
                            Debug.Print("Event: Request for Bootloader Patch");
                            this.SendAsync(Command.PatchBuilder(PatchType.HCI_EVENT_PATCHES_BOOTLOAD_REQ));
                            break;
                        case PatchType.HCI_EVENT_PATCHES_DRV_REQ:
                            Debug.Print("Event: Request for Driver Patch");
                            this.SendAsync(Command.PatchBuilder(PatchType.HCI_EVENT_PATCHES_DRV_REQ));
                            break;
                        case PatchType.HCI_EVENT_PATCHES_FW_REQ:
                            Debug.Print("Event: Request for Firmware Patch");
                            this.SendAsync(Command.PatchBuilder(PatchType.HCI_EVENT_PATCHES_FW_REQ));
                            break;
                    }
                    
                    break;
                case HciCommand.HCI_NETAPP_IPCONFIG:
                    this.PopulateNetworkProperties(aResponse);

                    //raise a wifi connected event with the SSID in the eventargs
                    if (!SuppressConnectionEvent) FireConnectedEvent();

                    break;
                case HciCommand.HCI_CMND_WLAN_IOCTL_SET_CONNECTION_POLICY:
                    Debug.Print("Response Received -- Connection Policy Set");
                    break;
                case HciCommand.HCI_NETAPP_DHCP:
                    Debug.Print("Response Received -- DHCP Configured. "); break; 
                case HciCommand.HCI_EVNT_WLAN_CONNECT:

                    if (aResponse.ToInt(5) != 0)
                    {
                        Debug.Print("Connection Failed, check your access point settings");  //TODO: this needs to be an event or part of the ConnectedEvent
                    }
                    else
                    {
                        this.IsConnected = true;
                        //WatchDogTimer.Change(5000, 2000);
                    }
                    break;
                case HciCommand.HCI_EVNT_WLAN_DISCONNECT:
                    this.IsConnected = false;
                    ClearNetworkProperties();
                    //WatchDogTimer.Change(Timeout.Infinite, Timeout.Infinite);
                    Debug.Print("Connection Disconnect."); break;
                case HciCommand.HCI_CMND_SOCKET:
                    Debug.Print("Acknowledged Socket Command: " + aResponse.ToHexString()); break;
                case HciCommand.HCI_CMND_BIND:
                    Debug.Print("Acknowledged Bind " + (aResponse.ToInt(5) == 0 ? "Success" : "Failed")); break;
                
                case HciCommand.HCI_CMND_RECVFROM:
                case HciCommand.HCI_CMND_RECV:
                    // example response: "040D100D00000000000500000001000000"
                    // example of inactive response: "0404100D0001000000C7FFFFFF00000000"
                    //WatchDogTimer.Change(2000, 2000);
                    
                    if ((int)aResponse.ToInt(9) == -57)
                    {
                        Debug.Print("RECV Command Response -- Socket " + aResponse.ToInt(5) + " -- INACTIVE!  Close?");
                        try
                        {
                            if (sockets[aResponse.ToInt(5)] != null && !sockets[aResponse.ToInt(5)].IsClosed)
                            {
                                sockets[aResponse.ToInt(5)].Close();
                            }
                            //sockets[aResponse.ToInt(5)]
                            //if (sockets[aResponse.ToInt(5)] != null) sockets[aResponse.ToInt(5)].Close();

                            if (Listeners[aResponse.ToInt(5)] != null && Listeners[aResponse.ToInt(5)] is UdpSocket) (Listeners[aResponse.ToInt(5)] as UdpSocket).Reset();
                            if (Listeners[aResponse.ToInt(5)] != null && Listeners[aResponse.ToInt(5)] is TcpSocket) (Listeners[aResponse.ToInt(5)] as TcpSocket).Reset();
                        }
                        catch { }

                        break;
                    }

                    Debug.Print("RECV Command Response on socket " + aResponse.ToInt(5) + ".  Bytes pending: " + aResponse.ToInt(9) + ", Flags: " + aResponse.ToInt(13)); break;
                case HciCommand.HCI_DATA_RECV:
                case HciCommand.HCI_DATA_RECVFROM:
                    Debug.Print("Data Received -- Socket Message Received");
                    //WatchDogTimer.Change(2000, 2000);

                    if (aResponse.Length < 24) 
                    {
                        Debug.Print("Invalid response payload received");
                        break;
                    }

                    socketID = (int)Utility.ExtractValueFromArray(aResponse, 5, 4);    
                    

                    if (socketID >= MaxSockets) break;

                    this.sockets[socketID].FireReceivedEvent(aResponse);
                    
                    break;
                case HciCommand.HCI_CMND_CLOSE_SOCKET:
                    Debug.Print("Socket Close Command: " + aResponse.ToInt(5)); break;
                case HciCommand.HCI_EVNT_BSD_TCP_CLOSE_WAIT:
                    Debug.Print("Socket Closed Event: " + aResponse.ToInt(5));
                    try
                    {
                        //this.sockets[aResponse.ToInt(5)].Close();
                        //if (!this.sockets[aResponse.ToInt(5)].IsClosed)
                            this.sockets[aResponse.ToInt(5)].Close();
                        //this.sockets[aResponse.ToInt(5)].IsClosed = true;
                    }
                    catch { }
                    
                    //this.sockets[aResponse.ToInt(5)].IsClosed = true;
                    break;
                case HciCommand.HCI_CMND_ACCEPT:
                    Debug.Print("Accept Response " + aResponse.ToHexString());
                    //WatchDogTimer.Change(2000, 2000);

                    // http://processors.wiki.ti.com/index.php/Non_blocking_accept
                    var status = (int)Utility.ExtractValueFromArray(aResponse, 9, 4);
                    var sourcePort = new byte[] { aResponse[15], aResponse[16] }.ToShort();
                    var sourceIP = new byte[] { aResponse[20], aResponse[19], aResponse[18], aResponse[17] };
                    
                    socketID = (int)Utility.ExtractValueFromArray(aResponse, 5, 4);

                    Debug.Print("Socket " + socketID + " is " + (status < 0 ? (status == -2 ? "Pending" : "Error") : "OK, Accepted from IP " + sourceIP.ToAddress() + ":" + sourcePort));
                
                    // special case for accept
                    //ushort socketAcceptOpcode = (ushort)(0x9000 + socketID);
                    //if (WaiterOpCodes.Contains(socketAcceptOpcode)) opcode = socketAcceptOpcode;

                    //if (status < 0) break;  // negative status value is error or pending

                    // Since status is greater than 0, it actually contains the new socket ID!  
                    // "new connection socket descriptor (>=0)- once TCP connection has been established. "

                    //var s = new Socket(this, status);

                    //if (sockets[socketID].ChildSocket != null && sockets[socketID].ChildSocket.IsClosed == false)
                    //{
                    //    Debug.Print("Whoa!  We just had 2 overlapping Accepts! ");
                    //    sockets[socketID].ChildSocket.Close();
                    //}
                    //else
                    //{
                    //    sockets[socketID].ChildSocket = s;
                    //    Socket.FireSocketAcceptedEvent(s, aResponse);
                    //}

                    break;
                case HciCommand.HCI_EVNT_SENDTO:
                case HciCommand.HCI_EVNT_SEND:
                    //WatchDogTimer.Change(2000, 2000);
                    if (aResponse != null && aResponse.Length >= 13)
                    {
                        socketID = (int)aResponse.ToInt(5);
                        var byteCount = (int)aResponse.ToInt(9);

                        if (byteCount == -57)
                        {
                            Debug.Print("SEND Command Response indicates socket " + aResponse.ToInt(5) + " was closed.  ");
                            try
                            {
                                if (sockets[aResponse.ToInt(5)] != null && !sockets[aResponse.ToInt(5)].IsClosed) sockets[aResponse.ToInt(5)].Close();
                                //sockets[aResponse.ToInt(5)]
                                //if (sockets[aResponse.ToInt(5)] != null) sockets[aResponse.ToInt(5)].Close();
                            }
                            catch { }

                            break;
                        }

                        Debug.Print("Sent " + byteCount + " bytes on Socket " + socketID);
                        if (byteCount > 0) sockets[socketID].LastActivity = PowerState.Uptime; 
                    }
                    else
                    {
                        Debug.Print("Unknown Send response, Payload: " + aResponse.ToHexString());
                    }
                    break;
                case HciCommand.HCI_EVNT_WLAN_UNSOL_DISCONNECT:
                    this.IsConnected = false;
                    ClearNetworkProperties();                    
                    Debug.Print("Disconnected."); break;
                case HciCommand.HCI_EVNT_DATA_UNSOL_FREE_BUFF:
                    if (aResponse != null && aResponse.Length >= 11)
                    {
                        Debug.Print("Handles: " + aResponse.ToShort(5) + ", buffers: " + aResponse.ToShort(7) + ", Released Packets: " + aResponse.ToShort(9));
                    }
                    else
                    {
                        Debug.Print("Unsolicited free buffer status update received. ");
                    }
                    break;
                case HciCommand.HCI_EVNT_WLAN_KEEPALIVE:
                   // WatchDogTimer.Change(2000, 2000);
                    Debug.Print("Keep Alive");
                    return;
                default:
                    Debug.Print("Handling Response for OpCode: " + opcode.ToHexString() + ", Payload: " + aResponse.ToHexString());
                    break;
            }

            //if (WaiterOpCodes.Contains(opcode))
            //{
            //    //Debug.Print("ResponseHandler: Clearing responses, setting waiter");
            //    responses.Clear();  //TODO: delete only old responses when count > 5
            //    responses.Add(opcode, aResponse);


            //    //if (WaiterOpCodes.Contains(opcode)) WaiterOpCodes.Remove(opcode);
            //    Waiter.Set();
            //}

        }

        private Thread ConnectedThread;

        private void FireConnectedEvent()
        {
            if (this.OnConnectedEvent != null)
            {
                // This is broken out into a seperate thread because if the user runs a loop or something in the Connected Event
                // Handler, the ResponseHandler polling will stop because the Invoke() will block until the Connected Event returns...

                if (ConnectedThread != null && ConnectedThread.IsAlive) ConnectedThread.Abort();
                ConnectedThread = new Thread(ConnectedInvoke);                
                ConnectedThread.Start();
            }
        }

        void ConnectedInvoke()
        {
            OnConnectedEvent.Invoke(null);
        }


        public string Name = string.Empty;
        internal byte[] MacAddress;

        public bool IsConnected { get; private set; }

        internal void ActivateSmartConfig(int timeout)
        {
            throw new NotImplementedException();

            //            // Reset all the previous configuration
            //wlan_ioctl_set_connection_policy(DISABLE, DISABLE, DISABLE);	
            //wlan_ioctl_del_profile(255);
	
            ////Wait until CC3000 is disconnected
            //while (ulCC3000Connected == 1)
            //{
            //    SysCtlDelay(100);
            //    hci_unsolicited_event_handler();
            //}
	
            //// Start blinking LED1 during Smart Configuration process
            //turnLedOn(1);
	
            //wlan_smart_config_set_prefix((char*)aucCC3000_prefix);
            //turnLedOff(1);		
	
            //// Start the SmartConfig start process
            //wlan_smart_config_start(1);
            //turnLedOn(1);       
	
            //// Wait for Smart config to finish
            //while (ulSmartConfigFinished == 0)
            //{
            //    turnLedOff(1);
            //    SysCtlDelay(16500000);
            //    turnLedOn(1);		  
            //    SysCtlDelay(16500000);
            //}
            //turnLedOn(1);	
            //// create new entry for AES encryption key
            //nvmem_create_entry(NVMEM_AES128_KEY_FILEID,16);
	
            //// write AES key to NVMEM
            //aes_write_key((unsigned char *)(&smartconfigkey[0]));
	
            //// Decrypt configuration information and add profile
            //wlan_smart_config_process();
	
            //// Configure to connect automatically to the AP retrieved in the 
            //// Smart config process
            //wlan_ioctl_set_connection_policy(DISABLE, DISABLE, ENABLE);
	
	
            //// reset the CC3000
            //wlan_stop();
	
            //DispatcherUartSendPacket((unsigned char*)pucUARTCommandSmartConfigDoneString,
            //                                                 sizeof(pucUARTCommandSmartConfigDoneString));
	
            //SysCtlDelay(100000);
            //wlan_start(0);	
	
            //// Mask out all non-required events
            //wlan_set_event_mask(HCI_EVNT_WLAN_KEEPALIVE|HCI_EVNT_WLAN_UNSOL_INIT|
            //                                        HCI_EVNT_WLAN_ASYNC_PING_REPORT);
        }


    public class PendingCommand : IMessagePacket
    {
        public PendingCommand()
        {
            CreateTime = PowerState.Uptime;
        }

        public byte[] Content { get; set; }
        public ushort WaitOpcode { get; set; }
        public TimeSpan CreateTime { get; set; }
    }

    public class ReceivedResponse : IMessagePacket
    {
        public ReceivedResponse()
        {
            CreateTime = PowerState.Uptime;
        }

        public byte[] Content { get; set; }
        public ushort WaitOpcode { get; set; }
        public TimeSpan CreateTime { get; set; }
    }

}





    //internal static class CC3000_Extensions
    //{
    //    internal static void Send(this CC3000Driver driver, Command theCommand)
    //    {
    //        lock (driver)
    //        {

    //            //wait for the signal that its ready
    //            //TODO: support a timeout
    //            while (driver.irq.Read()) { }

    //            //assert cs and (redundantly) wait for irq
    //            driver.cs.Write(false);

    //            int chunkSize = driver.IsInitialized ? 256 : 4;  // the initial chunksize must be 4 according to the docs
    //            int transmitSize = 0;
    //            int bytesTrasmitted = 0;
    //            byte[] commandToTransmit = theCommand;

    //            byte[] response = new byte[chunkSize];


    //            while (bytesTrasmitted < commandToTransmit.Length)
    //            {
    //                transmitSize = System.Math.Min(chunkSize, commandToTransmit.Length - bytesTrasmitted);
                    
    //                var msg = Utility.ExtractRangeFromArray(commandToTransmit, bytesTrasmitted, transmitSize);

    //                //write header and read the length bytes
    //                //driver.spi.WriteRead(msg, 0, transmitSize, response, 0, transmitSize, 0);
    //                driver.spi.Write(msg);

    //                driver.IsInitialized = true;
    //                bytesTrasmitted += transmitSize;
    //                chunkSize = 256;
    //            }

    //            driver.cs.Write(true);
    //            //negate cs

    //            //return new Response(response);
    //        }
    //    }


    //}




////*****************************************************************************
////
//// Values that can be used as HCI Commands and HCI Packet header defines
////
////*****************************************************************************
    public enum HciCommandType : byte
    {
        HCI_TYPE_CMND = 0x1,
        HCI_TYPE_DATA = 0x2,
        HCI_TYPE_PATCH = 0x3,
        HCI_TYPE_EVNT = 0x4
    }


    public enum PatchType : byte
    {
        HCI_EVENT_PATCHES_DRV_REQ = 0x1,
        HCI_EVENT_PATCHES_FW_REQ = 0x2,
        HCI_EVENT_PATCHES_BOOTLOAD_REQ = 0x3
    }

    /// <summary>
    /// WLAN constants
    /// </summary>
    internal class Constants
    {
        internal const int WLAN_SL_INIT_START_PARAMS_LEN         = (1)  ;
        internal const int WLAN_PATCH_PARAMS_LENGTH              = (8)  ;
        internal const int WLAN_SET_CONNECTION_POLICY_PARAMS_LEN = (12) ;
        internal const int WLAN_DEL_PROFILE_PARAMS_LEN           = (4)  ;
        internal const int WLAN_SET_MASK_PARAMS_LEN              = (4)  ;
        internal const int WLAN_SET_SCAN_PARAMS_LEN              = (100);
        internal const int WLAN_GET_SCAN_RESULTS_PARAMS_LEN      = (4)  ;
        internal const int WLAN_ADD_PROFILE_NOSEC_PARAM_LEN      = (24) ;
        internal const int WLAN_ADD_PROFILE_WEP_PARAM_LEN        = (36) ;
        internal const int WLAN_ADD_PROFILE_WPA_PARAM_LEN        = (44) ;
        internal const int WLAN_CONNECT_PARAM_LEN                = (29) ;
        internal const int WLAN_SMART_CONFIG_START_PARAMS_LEN    = (4)  ;

        internal const int SPI_HEADER_SIZE                           =  (5);
        internal const int SIMPLE_LINK_HCI_CMND_HEADER_SIZE          =  (4);
        internal const int HEADERS_SIZE_CMD = (SPI_HEADER_SIZE + SIMPLE_LINK_HCI_CMND_HEADER_SIZE);
        internal const int SIMPLE_LINK_HCI_DATA_CMND_HEADER_SIZE     =  (5);
        internal const int SIMPLE_LINK_HCI_DATA_HEADER_SIZE          =  (5);
        internal const int SIMPLE_LINK_HCI_PATCH_HEADER_SIZE         =  (2);

        internal const byte SL_PATCHES_REQUEST_DEFAULT      =(0); 
        internal const byte SL_PATCHES_REQUEST_FORCE_HOST   =(1);
        internal const byte SL_PATCHES_REQUEST_FORCE_NONE   =(2);

        internal const int SL_SET_SCAN_PARAMS_INTERVAL_LIST_SIZE   =(16);
        internal const int SL_SIMPLE_CONFIG_PREFIX_LENGTH          = (3);
        internal const int ETH_ALEN                                = (6);
        internal const int MAXIMAL_SSID_LENGTH                     = (32);

        /* NVMEM file ID - system files*/
        internal const int NVMEM_NVS_FILEID                           =(0) ;
        internal const int NVMEM_NVS_SHADOW_FILEID                    =(1) ;
        internal const int NVMEM_WLAN_CONFIG_FILEID                   =(2) ;
        internal const int NVMEM_WLAN_CONFIG_SHADOW_FILEID            =(3) ;
        internal const int NVMEM_WLAN_DRIVER_SP_FILEID                =(4) ;
        internal const int NVMEM_WLAN_FW_SP_FILEID                    =(5) ;
        internal const int NVMEM_MAC_FILEID                           =(6) ;
        internal const int NVMEM_FRONTEND_VARS_FILEID                 =(7) ;
        internal const int NVMEM_IP_CONFIG_FILEID                     =(8) ;
        internal const int NVMEM_IP_CONFIG_SHADOW_FILEID              =(9) ;
        internal const int NVMEM_BOOTLOADER_SP_FILEID                 =(10);
        internal const int NVMEM_RM_FILEID                            =(11);

        // file ID - user files                 
        internal const int NVMEM_AES128_KEY_FILEID                    =(12);
        internal const int NVMEM_SHARED_MEM_FILEID                    =(13);
        internal const int NVMEM_USER_FILE_1_FILEID                   =(14);

        // invalid nvmem           
        internal const int NVMEM_MAX_ENTRY                            = (16);
    }


    //WLAN_SET_MASK_PARAMS_LEN

    public enum HciCommand : ushort
    {
        UNKNOWN = 0x0FFF,
        HCI_CMND_WLAN_BASE = (0x0000),
        HCI_CMND_WLAN_CONNECT = 0x0001,
        HCI_CMND_WLAN_DISCONNECT = 0x0002,
        HCI_CMND_WLAN_IOCTL_SET_SCANPARAM = 0x0003,
        HCI_CMND_WLAN_IOCTL_SET_CONNECTION_POLICY = 0x0004,
        HCI_CMND_WLAN_IOCTL_ADD_PROFILE = 0x0005,
        HCI_CMND_WLAN_IOCTL_DEL_PROFILE = 0x0006,
        HCI_CMND_WLAN_IOCTL_GET_SCAN_RESULTS = 0x0007,
        HCI_CMND_EVENT_MASK = 0x0008,
        HCI_CMND_WLAN_IOCTL_STATUSGET = 0x0009,
        HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_START = 0x000A,
        HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_STOP = 0x000B,
        HCI_CMND_WLAN_IOCTL_SIMPLE_CONFIG_SET_PREFIX = 0x000C,
        HCI_CMND_WLAN_CONFIGURE_PATCH = 0x000D,

        HCI_CMND_SOCKET_BASE = 0x1000,
        HCI_CMND_SOCKET = 0x1001,
        HCI_CMND_BIND = 0x1002,
        HCI_CMND_RECV = 0x1004,
        HCI_CMND_ACCEPT = 0x1005,
        HCI_CMND_LISTEN = 0x1006,
        HCI_CMND_CONNECT = 0x1007,
        HCI_CMND_BSD_SELECT = 0x1008,
        HCI_CMND_SETSOCKOPT = 0x1009,
        HCI_CMND_GETSOCKOPT = 0x100A,
        HCI_CMND_CLOSE_SOCKET = 0x100B,
        HCI_CMND_RECVFROM = 0x100D,
        HCI_CMND_GETHOSTNAME = 0x1010,
        HCI_CMND_MDNS_ADVERTISE = 0x1011,

        //#define HCI_DATA_BASE                               0x80

        HCI_CMND_SEND = 0x0081,  // arg size byte and opcode byte combined
        HCI_CMND_SENDTO = 0x0083, //arg size byte and opcode byte combined

        HCI_DATA_RECV = 0x1885,  // arg size byte and opcode byte combined
        HCI_DATA_RECVFROM = 0x1884, //arg size byte and opcode byte combined

        //#define HCI_DATA_BSD_RECVFROM                       (0x04 + HCI_DATA_BASE)
        //#define HCI_DATA_BSD_RECV                           (0x05 + HCI_DATA_BASE)

        HCI_CMND_NVMEM_CBASE = (0x0200),

        HCI_CMND_NVMEM_CREATE_ENTRY = (0x0203),
        HCI_CMND_NVMEM_SWAP_ENTRY = (0x0205),
        HCI_CMND_NVMEM_READ = (0x0201),
        HCI_CMND_NVMEM_WRITE = (0x0090),
        HCI_CMND_NVMEM_WRITE_PATCH = (0x0204),
        HCI_CMND_READ_SP_VERSION = (0x0207),

        HCI_CMND_READ_BUFFER_SIZE = 0x400B,
        HCI_CMND_SIMPLE_LINK_START = 0x4000,

        HCI_CMND_NETAPP_BASE = 0x2000,

        HCI_NETAPP_DHCP = (0x0001 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_PING_SEND = (0x0002 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_PING_REPORT = (0x0003 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_PING_STOP = (0x0004 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_IPCONFIG = (0x0005 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_ARP_FLUSH = (0x0006 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_SET_DEBUG_LEVEL = (0x0008 + HCI_CMND_NETAPP_BASE),
        HCI_NETAPP_SET_TIMERS = (0x0009 + HCI_CMND_NETAPP_BASE),

        ////*****************************************************************************
        ////
        //// Values that can be used as HCI Events defines
        ////
        ////*****************************************************************************
        HCI_EVNT_WLAN_BASE = 0x0000,
        HCI_EVNT_WLAN_CONNECT = 0x0001,
        HCI_EVNT_WLAN_DISCONNECT = 0x0002,
        HCI_EVNT_WLAN_IOCTL_ADD_PROFILE = 0x0005,

        //#define  HCI_EVNT_SOCKET              HCI_CMND_SOCKET
        //#define  HCI_EVNT_BIND                HCI_CMND_BIND
        //#define  HCI_EVNT_RECV                HCI_CMND_RECV
        //#define  HCI_EVNT_ACCEPT              HCI_CMND_ACCEPT
        //#define  HCI_EVNT_LISTEN              HCI_CMND_LISTEN
        //#define  HCI_EVNT_CONNECT             HCI_CMND_CONNECT
        //#define  HCI_EVNT_SELECT              HCI_CMND_BSD_SELECT
        //#define  HCI_EVNT_CLOSE_SOCKET        HCI_CMND_CLOSE_SOCKET
        //#define  HCI_EVNT_RECVFROM            HCI_CMND_RECVFROM
        //#define  HCI_EVNT_SETSOCKOPT          HCI_CMND_SETSOCKOPT
        //#define  HCI_EVNT_GETSOCKOPT          HCI_CMND_GETSOCKOPT
        //#define  HCI_EVNT_BSD_GETHOSTBYNAME   HCI_CMND_GETHOSTNAME
        //#define  HCI_EVNT_MDNS_ADVERTISE   HCI_CMND_MDNS_ADVERTISE

        HCI_EVNT_SEND = 0x1003,
        //#define  HCI_EVNT_WRITE         0x100E
        HCI_EVNT_SENDTO = 0x100F,

        HCI_EVNT_PATCHES_REQ = 0x1000,

        HCI_EVNT_UNSOL_BASE = 0x4000,

        HCI_EVNT_WLAN_UNSOL_BASE = 0x8000,

        HCI_EVNT_WLAN_UNSOL_CONNECT = (0x0001 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_UNSOL_DISCONNECT = (0x0002 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_UNSOL_INIT = (0x0004 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_TX_COMPLETE = (0x0008 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_UNSOL_DHCP = (0x0010 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_ASYNC_PING_REPORT = (0x0040 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_ASYNC_SIMPLE_CONFIG_DONE = (0x0080 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_WLAN_KEEPALIVE = (0x0200 + HCI_EVNT_WLAN_UNSOL_BASE),
        HCI_EVNT_BSD_TCP_CLOSE_WAIT = (0x0800 + HCI_EVNT_WLAN_UNSOL_BASE),

        HCI_EVNT_DATA_UNSOL_FREE_BUFF = 0x4100,

        //#define HCI_EVNT_NVMEM_CREATE_ENTRY \
        //                                HCI_CMND_NVMEM_CREATE_ENTRY
        //#define HCI_EVNT_NVMEM_SWAP_ENTRY HCI_CMND_NVMEM_SWAP_ENTRY

        HCI_EVNT_NVMEM_READ = HCI_CMND_NVMEM_READ,
        HCI_EVNT_NVMEM_WRITE = 0x0202,

        HCI_EVNT_READ_SP_VERSION = HCI_CMND_READ_SP_VERSION,

        //#define  HCI_EVNT_INPROGRESS    0xFFFF


        //#define HCI_DATA_RECVFROM       0x84
        //#define HCI_DATA_RECV           0x85
        HCI_DATA_NVMEM = 0x1891,//0x91

        //#define HCI_EVENT_CC3000_CAN_SHUT_DOWN 0x99

        ////*****************************************************************************
        ////
        //// Prototypes for the structures for APIs.
        ////
        ////*****************************************************************************
        //#define HCI_DATA_HEADER_SIZE        (5)
        //#define HCI_EVENT_HEADER_SIZE       (5)
        //#define HCI_DATA_CMD_HEADER_SIZE    (5)
        //#define HCI_PATCH_HEADER_SIZE       (6)

        //#define HCI_PACKET_TYPE_OFFSET      (0)
        //#define HCI_PACKET_ARGSIZE_OFFSET   (2)
        //#define HCI_PACKET_LENGTH_OFFSET    (3)


        //#define HCI_EVENT_OPCODE_OFFSET (1)
        //#define HCI_EVENT_LENGTH_OFFSET (3)
        //#define HCI_EVENT_STATUS_OFFSET (4)
        //#define HCI_DATA_LENGTH_OFFSET  (3)
        Socket0AcceptEvent = 0x9000,
        Socket1AcceptEvent = 0x9001,
        Socket2AcceptEvent = 0x9002,
        Socket3AcceptEvent = 0x9003,

        Socket0RecvEvent = 0x9010,
        Socket1RecvEvent = 0x9011,
        Socket2RecvEvent = 0x9012,
        Socket3RecvEvent = 0x9013

    }
}
