// This file is part of WifiCC3000 - the Managed Wifi Driver for the TI CC3000
// Hosted on Codeplex: http://cc3000.codeplex.com
// WifiCC3000 is free software licensed under the Apache License 2.0
//  Copyright 2012 ValkyrieTech, LLC

using System;
using Microsoft.SPOT;
using System.IO;
using System.Text;
using Microsoft.SPOT.Hardware;
using System.Collections;

namespace Networking
{
    public class HttpResponse
    {
        // HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nCache-Control: no-cache; charset=utf-8\r\nContent-Length: 82\r\nConnection: close
        // Send the header
        public string ContentType;
        public string Connection;

        /// <summary>
        /// A string containing both the status code number and the Message.  Such as "HTTP/1.1 200 OK" or "HTTP/1.1 404 Not Found".  
        /// </summary>
        public string Status;
        public string Message;
        public string Header;

        /// <summary>
        /// Additional HTTP Headers in "key: value" format, such as "Cache-Control: no-cache; charset=utf-8"
        /// </summary>
        public Hashtable Headers = new Hashtable();
        internal Stream _content;

        /// <summary>
        /// Creates a new Http Response object.  
        /// </summary>
        /// <param name="body">Be aware of memory limitations here.  Use small strings with this constructor.  </param>
        /// <param name="contentType"></param>
        /// <param name="status"></param>
        /// <param name="connection"></param>
        public HttpResponse(string body, string contentType = "text/html", string status = "200 OK", string connection = "close")
        {
            ContentType = contentType;
            Connection = connection;
            Status = status;

            _content = new System.IO.MemoryStream(Encoding.UTF8.GetBytes(body));
        }

        public HttpResponse(Stream content, string contentType = "text/html", string status = "200 OK", string connection = "close")
        {
            ContentType = contentType;
            Connection = connection;
            Status = status;

            _content = content;
        }

        /// <summary>
        /// Creates a new Response object by parsing the content into the Response object properties
        /// </summary>
        /// <param name="content"></param>
        public HttpResponse(byte[] content, bool omitContent = false)
        {
            int contentStart = content.Locate(new byte[4] { 0x0d, 0x0a, 0x0d, 0x0a });
            int headerStart = content.Locate(new byte[2] { 0x0d, 0x0a });

            Header = string.Empty;
            Message = string.Empty;

            Status = new string(Encoding.UTF8.GetChars(Utility.ExtractRangeFromArray(content, 0, headerStart)));
            if ((contentStart > (headerStart + 2)) && headerStart > 5)
            {
                int colonLocation = -1;
                int start = headerStart + 2;
                bool malformed = false;

                // Parse all the header keys and values (Note: this is intentionally parsed directly from the byte array to minimize creating and destroying strings that consume memory and need garbage collection)
                for (int i = headerStart + 2; i <= contentStart; i++)
                {
                    if (content[i] == 0x3A && colonLocation == -1) colonLocation = i;
                    if (content[i] > 0x7E || content[i] < 0x09) malformed = true;

                    if (content[i] == 0x0D || content[i] == 0x0A)
                    {
                        if (colonLocation > start && !malformed)
                        {
                            // By handling the exception at each parameter, we can salvage the other headers if one contains an invalid character
                            // Although, we should be avoiding any exceptions with the malformed flag.  
                            try
                            {
                                Headers.Add(new string(Encoding.UTF8.GetChars(content, start, colonLocation - start)).Trim(),
                                            new string(Encoding.UTF8.GetChars(content, colonLocation + 1, i - colonLocation)).Trim());
                            }
                            catch { }
                        }

                        colonLocation = -1;
                        start = i + 1;
                        malformed = false;
                    }
                }

                //Header = new string(Encoding.UTF8.GetChars(Utility.ExtractRangeFromArray(content, headerStart + 2, contentStart - (headerStart + 2))));
                if (!omitContent && contentStart + 4 < content.Length)
                    Message = new string(Encoding.UTF8.GetChars(Utility.ExtractRangeFromArray(content, contentStart + 4, content.Length - (contentStart + 4))));
            }


        }


        public byte[] HeaderSection
        {
            get
            {
                string header = "HTTP/1.1 " + Status + (ContentType != null && ContentType != string.Empty ? ("\r\nContent-Type: " + ContentType) : string.Empty);

                foreach (string headerKey in Headers.Keys)
                {
                    header += "\r\n" + headerKey + ": " + (Headers[headerKey] as string);
                }

                header += ((_content != null && _content.Length > 0) ? ("\r\nContent-Length: " + _content.Length.ToString()) : string.Empty)
                          + (Connection != null && Connection != string.Empty ? ("\r\nConnection: " + Connection) : string.Empty)
                          + "\r\n\r\n";

                return Encoding.UTF8.GetBytes(header);
            }
        }



    }

}
