// This file is part of WifiCC3000 - the Managed Wifi Driver for the TI CC3000
// Hosted on Codeplex: http://cc3000.codeplex.com
// WifiCC3000 is free software licensed under the Apache License 2.0
//  Copyright 2012 ValkyrieTech, LLC

using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using System.Threading;

namespace Networking
{
    public class UdpSocket : IDisposable
    {
        Socket baseSocket;
        Thread listener;

        public delegate void UdpReceivedEventHandler(UdpSocket sender, byte[] message, byte[] remoteIP, ushort remotePort);
        public event UdpReceivedEventHandler OnDataReceived;

        ushort listenPort = 0;
        byte[] listenAddress = null;
        //private bool IsResetting = false;

        public UdpSocket(CC3000Driver driver)
        {
            baseSocket = new Socket(driver, (int)AddressFamily.INET, (int)SocketType.DGRAM, (int)Protocol.UDP);
        }

        private UdpSocket(Socket s)
        {
            baseSocket = s;
        }

        //public void Listen(ushort port)
        //{
        //    //baseSocket.Bind(port);
        //    StopListening();

        //    if (listener != null && listener.IsAlive) listener.Abort();

        //    listenPort = port;
        //    listenAddress = null;

        //    this.Bind(null, port);

        //    listener = new Thread(this.Listen);
        //    listener.Start();

        //}

        public void Listen(ushort port, byte[] address = null)
        {
            //baseSocket.Bind(port);
            StopListening();

            if (listener != null && listener.IsAlive) listener.Abort();

            listenPort = port;
            listenAddress = address;

            this.Bind(address, port);

            if (baseSocket.wifi.Listeners.Contains(baseSocket.socketID))
                baseSocket.wifi.Listeners[baseSocket.socketID] = this;
            else
                baseSocket.wifi.Listeners.Add(baseSocket.socketID, this);

            listener = new Thread(this.Listen);
            listener.Start();

        }

        public void StopListening()
        {
            listenPort = 0;  // This should allow the thread to exit

            //if (listener != null && listener.IsAlive) listener.Abort();
        }

        //new Thread(delegate
        //{
        //    try
        //    {
        //        Debug.WriteLine("Try DNS Lookup");

        //        result = Dns.GetHostEntry(HostnameOrAddress);

        //        Debug.WriteLine("DNS Looked up!");
        //    }
        //    catch { }
        //}).Start();

        /// <summary>
        /// Frequency of checking for Received Data (in milliseconds)
        /// </summary>
        public uint PollFrequency = 200;  


        private void Listen()
        {
            baseSocket.OnDataReceived += baseSocket_OnDataReceived;

            //baseSocket.Bind(listenPort);

            while (listenPort > 0)
            {
                // poll for newly received data...
                if (!baseSocket.wifi.PendingCommands.Any((ushort)(0x9020 + baseSocket.socketID))) baseSocket.Receive();  

                Thread.Sleep((int)PollFrequency);  // Note: this sleep may limit the bandwidth of data being received on UDP
            }

            baseSocket.OnDataReceived -= baseSocket_OnDataReceived;
        }

        void baseSocket_OnDataReceived(Socket sender, byte[] message, byte[] sourceIP, ushort sourcePort)
        {
            if (this.OnDataReceived != null && sender.socketID == baseSocket.socketID)
            {
                if (message.Length > 0) baseSocket.LastActivity = PowerState.Uptime; 
                this.OnDataReceived.Invoke(this, message, sourceIP, sourcePort);
            }
        }

        public void Send(byte[] message, byte[] destinationIP, ushort port)
        {
            baseSocket.Send(message, destinationIP, port);
        }


        public void Close()
        {
            if (baseSocket.wifi.Listeners.Contains(baseSocket.socketID)) baseSocket.wifi.Listeners.Remove(baseSocket.socketID);

            baseSocket.Close();
        }

        public void Dispose()
        {
            // Kill listener threads

            baseSocket.Close();
        }



        internal void Reset()
        {
            //this.IsResetting = true;

            if (listenPort > 0)
            {
                Debug.Print("Resetting Listener for port: " + listenPort);
                //if (baseSocket.ChildSocket != null) baseSocket.ChildSocket.IsClosed = true;

                //while (listener != null && listener.IsAlive) Thread.Sleep(50);

                //this.IsResetting = false;

                baseSocket = new Socket(baseSocket.wifi, (int)AddressFamily.INET, (int)SocketType.DGRAM, (int)Protocol.UDP);
                this.Listen(listenPort, listenAddress);
                Debug.Print("UDP Listener Restart complete.");
            }
        }

        public void Bind(byte[] address, ushort port)
        {
            baseSocket.Bind(address, port);
        }

        public void RecieveFrom()
        {
            baseSocket.Receive();

        }

    }

    public class TcpSocket : IDisposable
    {
        public delegate void TcpReceivedEventHandler(TcpSocket sender, byte[] message, byte[] remoteIP, ushort remotePort);
        public event TcpReceivedEventHandler OnDataReceived;

        Socket baseSocket;
        Thread listener;

        private bool IsResetting = false;

        ushort listenPort = 0;
        private ushort SocketTimeout = 10;

        public TcpSocket(CC3000Driver driver, ushort timeout = 10)
        {
            baseSocket = new Socket(driver, (int)AddressFamily.INET, (int)SocketType.STREAM, (int)Protocol.TCP);
            SocketTimeout = timeout;
        }

        private TcpSocket(Socket s, ushort timeout = 10)
        {
            baseSocket = s;
            SocketTimeout = timeout;
        }

        public void Listen(ushort port)
        {
            baseSocket.Bind(port);

            if (listener != null && listener.IsAlive) listener.Abort();

            listenPort = port;

            listener = new Thread(this.Listen);
            listener.Start();

        }

        public void StopListening()
        {
            listenPort = 0;  // This should allow the thread to exit

            //if (listener != null && listener.IsAlive) listener.Abort();
        }

        //new Thread(delegate
        //{
        //    try
        //    {
        //        Debug.WriteLine("Try DNS Lookup");

        //        result = Dns.GetHostEntry(HostnameOrAddress);

        //        Debug.WriteLine("DNS Looked up!");
        //    }
        //    catch { }
        //}).Start();


        private void Listen()
        {
            Socket.OnSocketAccepted += Socket_OnSocketAccepted;

            // Enable accept blocking mode
            //baseSocket.SetOption(SocketOption.AcceptBlockingMode, (1).ToBytesRev());
            
            baseSocket.Listen(listenPort, 5);

            if (baseSocket.wifi.Listeners.Contains(baseSocket.socketID))
                baseSocket.wifi.Listeners[baseSocket.socketID] = this;
            else
                baseSocket.wifi.Listeners.Add(baseSocket.socketID, this);

            //int acceptedSocket = -1;

            int status = 0;

            while (listenPort > 0)
            {
                // we have to poll with "Accept" to see if a connection request is pending
                // http://processors.wiki.ti.com/index.php/Non_blocking_accept

                status = -10;

                //if (acceptedSocket < 0 || sockets[acceptedSocket] == null)
                if (baseSocket.ChildSocket == null || baseSocket.ChildSocket.IsClosed)
                {
                    if (!baseSocket.wifi.IsResetting)
                    {
                        Debug.Print("Childsocket is null? " + (baseSocket.ChildSocket == null).ToString() + (baseSocket.ChildSocket != null ? ", ChildSocket is closed? " + baseSocket.ChildSocket.IsClosed.ToString() : ""));

                        var aResponse = baseSocket.Accept();

                        if (aResponse != null)
                        {
                            status = (int)Utility.ExtractValueFromArray(aResponse, 9, 4);
                            var sourcePort = new byte[] { aResponse[15], aResponse[16] }.ToShort();
                            var sourceIP = new byte[] { aResponse[20], aResponse[19], aResponse[18], aResponse[17] };

                            if (sourcePort != 0 && sourceIP.ToInt() != 0 && status >= 0)
                            {
                                // We have a good, open socket, let's poll for data while it's open...

                                var s = new Socket(baseSocket.wifi, status);

                                if (baseSocket.ChildSocket != null && baseSocket.ChildSocket.IsClosed == false)
                                {
                                    Debug.Print("Whoa!  We just had 2 overlapping Accepts! ");
                                    baseSocket.ChildSocket.Close();
                                }

                                baseSocket.ChildSocket = s;
                                s.SetOption(SocketOption.ReceiveBlockingMode, 0.ToBytesRev());                                
                                //                                Socket.FireSocketAcceptedEvent(s, aResponse);

                                s.OnDataReceived += s_OnDataReceived;

                                //childSocketID = baseSocket.socketID;

                                Thread.Sleep(50);  // CC3000 needs a little time to instantiate the socket? 

                                while (s != null && !s.IsClosed && !s.IsClosing && !baseSocket.wifi.IsResetting && (PowerState.Uptime.Subtract(s.LastActivity).Ticks < TimeSpan.TicksPerSecond * SocketTimeout))  
                                {
                                    // if we already have a pending RECV command in the queue, there is no need to add another...
                                    if (!s.wifi.PendingCommands.Any((ushort)(0x9010 + s.socketID))) s.TcpReceive(512, false);
                                    System.Threading.Thread.Sleep(50);
                                }

                                s.OnDataReceived -= s_OnDataReceived;

                                if (!s.IsClosed && !s.IsClosing && !baseSocket.wifi.IsResetting)
                                {
                                    Debug.Print("Last Activity on socket " + s.socketID + " was " + (PowerState.Uptime.Subtract(baseSocket.LastActivity).Ticks / TimeSpan.TicksPerSecond).ToString() + " seconds ago.");
                                    s.Close();
                                }
                            }
                        }
                        else
                        {
                            Debug.Print("Accept Response was null");
                        }
                    }
                    else
                    {
                        Thread.Sleep(50);
                    }




                    //if (response == null)
                    //    baseSocket.wifi.Send(Command.Builder(HciCommand.HCI_CMND_READ_BUFFER_SIZE));

                    //if (response != null && -57 == (int)Utility.ExtractValueFromArray(response, 9, 4))
                    //{
                    //    Debug.Print("RESTARTING LISTENER!!!");
                    //    // Reset the Listener!
                    //    baseSocket.Close();

                    //    baseSocket = baseSocket = new Socket(baseSocket.wifi, (int)AddressFamily.INET, (int)SocketType.STREAM, (int)Protocol.TCP);
                    //    baseSocket.Bind(listenPort);
                    //    baseSocket.Listen(listenPort, 5);
                    //    Debug.Print("Listener Restart complete.");
                    //}

                    //acceptedSocket = (int)Utility.ExtractValueFromArray(response, 5, 4);
                }

                // WOW!  This may be a critically sensitive value.  100ms bombs, but 200ms works... why?
                //System.Threading.Thread.Sleep(300);

                // if Pending (aka. no pending connections from clients), then sleep a while...
                if (status == -2)
                    Thread.Sleep(300);
                else
                    Thread.Sleep(20);

                //TODO: if we are seeing continuous -1 responses, that indicates an error in the socket, and we should try resetting the socket if that happens continuous.  
                // An occassional error is probably not a problem during intense bursts of messages...
            }

            Socket.OnSocketAccepted -= Socket_OnSocketAccepted;
        }

        void Socket_OnSocketAccepted(Socket sender, byte[] sourceIP, ushort sourcePort)
        {
            //if (sourcePort == 0 || sourceIP.ToInt() == 0) 
            //    return;

            //sender.OnDataReceived += s_OnDataReceived;

            //childSocketID = sender.socketID;

            //while (sender != null && !sender.IsClosed && !sender.IsClosing && !sender.IsSending)  //TODO: change the while loop so that it times out after 10 seconds if nothing new is received and returns the socket?
            //{
            //    sender.TcpReceive(512, false);
            //    System.Threading.Thread.Sleep(200);
            //}

            //sender.OnDataReceived -= s_OnDataReceived;
        }

        void s_OnDataReceived(Socket sender, byte[] message, byte[] sourceIP, ushort sourcePort)
        {
            Debug.Print("onDataReceived -- " + sender.socketID);

            sender.PendingReceiveBytes -= message.Length;

            if (this.OnDataReceived != null) // && childSocketID == sender.socketID)
            {
                if (message.Length > 0) baseSocket.LastActivity = PowerState.Uptime; 
                this.OnDataReceived.Invoke(new TcpSocket(sender), message, sourceIP, sourcePort);
            }
        }

        public void Reset()
        {
            this.IsResetting = true;

            if (listenPort > 0)
            {
                Debug.Print("Resetting Listener for port: " + listenPort);
                if (baseSocket.ChildSocket != null) baseSocket.ChildSocket.IsClosed = true;

                //while (listener != null && listener.IsAlive) Thread.Sleep(50);

                //this.IsResetting = false;
                
                baseSocket = new Socket(baseSocket.wifi, (int)AddressFamily.INET, (int)SocketType.STREAM, (int)Protocol.TCP);
                baseSocket.Bind(listenPort);
                baseSocket.Listen(listenPort, 5);
                Debug.Print("Listener Restart complete.");
            }
        }

        public void Send(byte[] message)
        {
            baseSocket.Send(message);
        }

        public void Close()
        {
            if (baseSocket.wifi.Listeners.Contains(baseSocket.socketID)) baseSocket.wifi.Listeners.Remove(baseSocket.socketID);

            baseSocket.Close();
        }

        public void Dispose()
        {
            // Kill listener threads

            baseSocket.Close();
        }
    }


    public class Socket
    {
        internal int socketID = -1;
        internal CC3000Driver wifi = null;

        internal ushort boundPort = 0;

        static byte[] packetTail = new byte[8] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

        public delegate void ReceivedEventHandler(Socket sender, byte[] message, byte[] sourceIP, ushort sourcePort);
        public event ReceivedEventHandler OnDataReceived;

        public delegate void SocketAcceptedEventHandler(Socket sender, byte[] sourceIP, ushort sourcePort);
        public static event SocketAcceptedEventHandler OnSocketAccepted;

        public Socket ChildSocket = null;

        public bool IsSending = false;
        internal TimeSpan LastActivity = PowerState.Uptime;

        private object lockObj = new object();

        private Protocol socketProtocol = Protocol.IP;

        /// <summary>
        /// Create and open a socket
        /// </summary>
        /// <param name="driver"></param>
        /// <param name="domain"></param>
        /// <param name="type"></param>
        /// <param name="protocol"></param>
        /// <param name="timeout">Time to wait for socket creation response in milliseconds</param>
        internal Socket(CC3000Driver driver, int domain, int type, int protocol, uint timeout = 10000)
        {
            wifi = driver;

            string protocolStr = "Unknown";

            switch (protocol)
            {
                case (int)Protocol.UDP: protocolStr = "UDP"; break;
                case (int)Protocol.TCP: protocolStr = "TCP"; break;
            }

            socketProtocol = (Protocol)protocol;

            Debug.Print("Opening " + protocolStr + " Socket");
            var r = driver.Send(Command.Builder(HciCommand.HCI_CMND_SOCKET, domain, type, protocol), timeout);

            if (r == null)
            {
                throw new Exception("Failed to open a new socket");
                
                //wifi.ResetAndReconnect();
                r = driver.Send(Command.Builder(HciCommand.HCI_CMND_SOCKET, domain, type, protocol), timeout);
            }

            socketID = r[5];

            // http://e2e.ti.com/support/low_power_rf/f/851/t/269806.aspx
            //this.SetOption(SocketOption.ReceiveTimeout, (200).ToBytesRev());
            this.SetOption(SocketOption.ReceiveBlockingMode, 0.ToBytesRev());

            if (socketID < wifi.sockets.Length)
                wifi.sockets[socketID] = this;
            else
                Debug.Print("Socket ID is out of expected range!");

            IsClosed = false;

            this.LastActivity = PowerState.Uptime;

            Debug.Print("Socket ID is " + socketID);
        }

        internal Socket(CC3000Driver cC3000Driver, int SocketID)
        {
            wifi = cC3000Driver;
            socketID = SocketID;

            if (socketID < wifi.sockets.Length)
                wifi.sockets[socketID] = this;
            else
                Debug.Print("Socket ID is out of expected range!");

            IsClosed = false;

            this.LastActivity = PowerState.Uptime;
            Debug.Print("Socket ID is " + socketID);
        }

        public void Send(byte[] data, uint retryCount = 3)
        {
            if (this.IsClosed || this.IsClosing) return;
            
            IsSending = true;

            Debug.Print("Sending " + data.Length + " bytes of data on socket " + socketID);

            byte[] result = null;

            //while (result == null && retryCount-- > 0)
            //{
            //    Debug.Print("Sending " + data.Length + " bytes of data on socket " + socketID);

                wifi.Send(Command.DataBuilder(HciCommand.HCI_CMND_SEND, data, null, socketID, 12, data.Length, 0), (ushort)HciCommand.HCI_EVNT_SEND, 1);
                //result = wifi.Send(Command.DataBuilder(HciCommand.HCI_CMND_SEND, data, null, socketID, 12, data.Length, 0), (ushort)HciCommand.HCI_EVNT_SEND, 1500);

            //}
            //wifi.Send(Command.DataBuilder(HciCommand.HCI_CMND_SEND, data, null, socketID, 12, data.Length, 0));
            //wifi.Send(Command.DataBuilder(HciCommand.HCI_CMND_SEND, data, null, socketID, 12, data.Length, 0));
            Debug.Print(DateTime.Now.ToLocalTime() + " - Send Data command dispatched - result: " + result == null ? "null" : result.ToHexString());

            LastSend = PowerState.Uptime;
            
            IsSending = false;
        }

        public void Send(byte[] data, byte[] destinationIP, ushort port)
        {
            if (this.IsClosed || this.IsClosing || data == null || destinationIP == null) return;

            IsSending = true;
            //Debug.Print("Send mem1: " + Networking.CC3000Driver.MemDelta(true));

//            Debug.Print("Sending Message on socket " + socketID + " to " + destinationIP.ToAddress() + ":" + port);

            // Assemble the address and port info as a byte array
            var addressBytes = new byte[16];  // I have no idea why this is not 8.  The address block seems to always have 8 extra bytes (usually 0x00s or junk)
            addressBytes[0] = 0x02;
            addressBytes.Overwrite(2, port.ToBytes());
            addressBytes.Overwrite(4, destinationIP);

            wifi.Send(Command.DataBuilder(HciCommand.HCI_CMND_SENDTO, data, addressBytes, socketID, 20, data.Length, 0, data.Length + 8, 8), (ushort)HciCommand.HCI_EVNT_SENDTO, 0);

              //  Debug.Print("Send mem2: " + Networking.CC3000Driver.MemDelta(true));

               // wifi.SendAsync(x);
      //          GC.WaitForPendingFinalizers();
                //GC.SuppressFinalize();
            
            //Debug.Print("Send mem3: " + Networking.CC3000Driver.MemDelta(true));


            #region Annotated sample SendTo command
            //01 00 33 00 00 - SPI Header, HCI Length of 0x33 or 51 decimal
            //02 - Write Type
            //83 - Opcode 0x83 HCI_CMND_SENDTO
            //18 - 0x18 Argument length? 24? For SendTo - SOCKET_SENDTO_PARAMS_LEN(24), for Send-HCI_CMND_SEND_ARG_LENGTH(16)
            //2D 00 - Payload Length? 45 decimal
            //00 00 00 00 - Socket ID of 0
            //14 00 00 00 - Arg size =  SOCKET_SENDTO_PARAMS_LEN(24) - Length of Socket ID (4) = 20 = 0x14; Note: SEND (as opposed to SENDTO) uses HCI_CMND_SEND_ARG_LENGTH(16) - Length of SocketID (4)
            //05 00 00 00 - Message length
            //00 00 00 00 - flags (all 0s)
            //0D 00 00 00 - SENDTO Only - Address offset? Calculated from Message Length(5) + (Size of Message Length(4) x 2) = 5 + (4 x 2) = 13 = 0x0D
            //08 00 00 00 - SENDTO Only - Address Length (always 8?  what about ipv6?  eek!) 
            //31 32 33 34 35 - THE DATA/MESSAGE (ASCII for "12345")
            //02 00 C7 38 - 0x0002 is family AF_INET (always and type is short), but 0xC738 is the port (SENDTO Only?)
            //C0 A8 01 54 - Destination IP 192.168.1.84 (SENDTO Only?)
            //F9 FF FF FF - ?? This seems to be the same everytime, no matter how the data content changes
            //CB 12 00 20 - ?? This is usually 4 bytes of 0x00.  I have no clue what this is...
            //00 - Padding byte
            #endregion

            IsSending = false;
        }

        //public void Receive(byte[] sourceIP, int sourcePort, bool blocking)
        //{
        //    // Parameters are all int -- SocketID, BufferLength, flags (blocking or not blocking)
        //    var x = Command.Builder(HciCommand.HCI_CMND_RECVFROM, socketID, 256, blocking ? 1 : 0);

        //    wifi.SendAsync(x);
        //}

        public void Receive(int chunkSize = 512, bool blocking = false)
        {
            if (boundPort == 0) throw new Exception("Port must be bound before trying to receive");
            if (wifi.IsResetting || this.IsClosed || this.IsClosing) return;
            // Parameters are all int -- SocketID, BufferLength, flags (blocking or not blocking)
            wifi.Send(Command.Builder(HciCommand.HCI_CMND_RECVFROM, socketID, chunkSize, blocking ? 1 : 0), (ushort)(0x9020 + socketID), 2000); //, (byte)0x02));
        }

        public byte[] ReceiveSync(int chunkSize = 512, bool blocking = false)
        {
            if (boundPort == 0) throw new Exception("Port must be bound before trying to receive");
            if (wifi.IsResetting || this.IsClosed || this.IsClosing) return null;
            // Parameters are all int -- SocketID, BufferLength, flags (blocking or not blocking)

            // Assemble the address and port info as a byte array
            //var addressBytes = new byte[16];  // I have no idea why this is not 8.  The address block seems to always have 8 extra bytes (usually 0x00s or junk)
            //addressBytes[0] = 0x02;
            //addressBytes.Overwrite(2, port.ToBytes());
            //addressBytes.Overwrite(4, destinationIP);

            return wifi.Send(Command.Builder(HciCommand.HCI_CMND_RECVFROM, socketID, chunkSize, blocking ? 1 : 0), (ushort)(0x9020 + socketID), 0); //, (byte)0x02));
        }

        internal int PendingReceiveBytes = 0;

        public void TcpReceive(int chunkSize = 512, bool blocking = false)
        {
            var deadSocket = false;

            if (deadSocket) this.Send(new byte[0]);
            
            // if (boundPort == 0) throw new Exception("Port must be bound before trying to receive");
            if (wifi.IsResetting || this.IsClosed || this.IsClosing || PendingReceiveBytes > 0 || IsSending || LastSend.Ticks > 0 ) return; //(PowerState.Uptime.Subtract(LastSend).Ticks < (TimeSpan.TicksPerSecond * 25))) return;
            // Parameters are all int -- SocketID, BufferLength, flags (blocking or not blocking)
            
            wifi.Send(Command.Builder(HciCommand.HCI_CMND_RECV, socketID, chunkSize, blocking ? 1 : 0), (ushort)(0x9010 + socketID), 0); //, (byte)0x02)).ToHexString());
            return;
            
            var aResponse = wifi.Send(Command.Builder(HciCommand.HCI_CMND_RECV, socketID, chunkSize, blocking ? 1 : 0), (ushort)(0x9010 + socketID), 2000); //, (byte)0x02)).ToHexString());
            //Debug.Print(wifi.Send(Command.Builder(HciCommand.HCI_CMND_RECV, socketID, chunkSize, blocking ? 1 : 0, (byte)0x02)).ToHexString());

            if (aResponse == null) return;

            if ((int)aResponse.ToInt(9) == -57)
            {
                Debug.Print("RECV Command Response -- Socket " + aResponse.ToInt(5) + " -- INACTIVE.");
                //this.IsClosed = true;
                return;
            }

            PendingReceiveBytes = (int)aResponse.ToInt(9);

            Debug.Print("RECV Command Response on socket " + aResponse.ToInt(5) + ".  Bytes pending: " + aResponse.ToInt(9) + ", Flags: " + aResponse.ToInt(13)); 

            Debug.Print("Tcp Receive");
        }

        public void Bind(ushort port)
        {
            this.Bind(wifi.IPAddress, port);
        }

        public void Bind(byte[] address, ushort port)
        {
            boundPort = 0; // reset

            if (port > 0)
            {                
                byte[] response;

                // Parameters are all int -- SocketID, UInteger 8, Address Length, Address family, port, ip address

                if (socketProtocol == Protocol.TCP || address == null)
                {
                    response = wifi.Send(Command.Builder(HciCommand.HCI_CMND_BIND, socketID, 8, 8, (ushort)2, port.ToBytes(), 0));

                    Debug.Print("Binding socket " + socketID + " to port " + port);
                }
                else
                {
                    // Assemble the address and port info as a byte array
                    var addressBytes = new byte[16];
                    addressBytes[0] = 0x02;
                    addressBytes.Overwrite(2, port.ToBytes());
                    addressBytes.Overwrite(4, address);

                    Debug.Print("Binding socket " + socketID + " to port " + port + " and address: " + address.ToAddress());

                    response = wifi.Send(Command.Builder(HciCommand.HCI_CMND_BIND, socketID, 8, addressBytes.Length, addressBytes));
                }

                if (response == null || (int)response.ToInt(5) != 0) throw new Exception("Failed to bind socket port " + port);
                boundPort = port;
            }
        }

        internal static void FireSocketAcceptedEvent(Socket s, byte[] message)
        {
            if (OnSocketAccepted != null)
                OnSocketAccepted.Invoke(s, new byte[4] { message[20], message[19], message[18], message[17] }, (new byte[] { message[15], message[16] }.ToShort()));
        }

        internal void FireReceivedEvent(byte[] message)
        {
            if (this.OnDataReceived != null)
            {
                int length = (int)Utility.ExtractValueFromArray(message, 13, 4);

                if (length > 0) LastActivity = PowerState.Uptime;

                this.OnDataReceived.Invoke(this, Utility.ExtractRangeFromArray(message, 29, length), Utility.ExtractRangeFromArray(message, 25, 4), (new byte[] { message[24], message[23] }.ToShort()));
            }
        }

        public byte[] Accept()
        {
            Debug.Print(DateTime.Now.ToLocalTime() + " - Starting Accept for socket " + socketID);

            return wifi.Send(Command.Builder(HciCommand.HCI_CMND_ACCEPT, socketID), (ushort)(0x9000 + socketID), 2000);
        }

        public void Listen(ushort port, uint backlog = 0)
        {
            //this.Bind(port);

            wifi.Send(Command.Builder(HciCommand.HCI_CMND_LISTEN, socketID, backlog));
        }

        public void Listen(uint backlog = 0)
        {
            if (boundPort == 0) return;

            wifi.Send(Command.Builder(HciCommand.HCI_CMND_LISTEN, socketID, backlog));
        }

        internal void SetOption(SocketOption option, byte[] value)
        {
            //TODO: Verify that this method works
            // Helpful info: http://e2e.ti.com/support/low_power_rf/f/851/t/208107.aspx
            // Parameters: socketID, socket level (always 0xFFFF or -1), option enum, 8, option length ???, value 
            wifi.Send(Command.Builder(HciCommand.HCI_CMND_SETSOCKOPT, socketID, -1, (int)option, 8, value.Length, value));

            //wifi.Send(Command.Builder(HciCommand.HCI_CMND_SETSOCKOPT, socketID, -1, 2, (int)option, 8, 8, value));
        }

        private byte[] GetOption(SocketOption option)
        {
            //TODO: Verify that this method works
            
            // Parameters: socketID, socket level (always 0xFFFF or -1), option enum  
            return wifi.Send(Command.Builder(HciCommand.HCI_CMND_GETSOCKOPT, socketID, -1, (int)option));
        }

        internal bool IsClosed = true;
        internal bool IsClosing = false;
        internal TimeSpan LastSend = new TimeSpan(0);

        
        /// <summary>
        /// Calling this Close method is essential.  Even if the Client disconnects, 
        /// this method must still be called to close the socket.  Otherwise, the cc3000
        /// will not reuse the socket ID and with only 4 total, you will quickly run out.  
        /// </summary>
        public void Close()
        {

            lock (lockObj)
            {

                if (IsClosed) return;

                //TODO: Verify that this method works
                IsClosing = true;

                Debug.Print("Sending Command to Close Socket " + socketID);

                if (IsClosed == false)
                {
                    var response = wifi.Send(Command.Builder(HciCommand.HCI_CMND_CLOSE_SOCKET, socketID));

                    //Debug.Assert(response == null, "Socket did not close!");
                    if (response == null) Debug.Print("Socket " + socketID + " did not close.");


                    IsClosed = true;
                }

                IsClosing = false;

                //try
                //{

                //    if (socketID < wifi.sockets.Length) wifi.sockets[socketID] = null;
                //}
                //catch
                //{
                //    Debug.Print("Bad socket id? ");
                //}
            }
        }

    }

    public enum SocketType : int
    {
        STREAM = 1,
        DGRAM = 2,
        RAW = 3,          // Raw sockets allow new IPv4 protocols to be implemented in user space. A raw socket receives or sends the raw datagram not including link level headers
        RDM = 4,
        SEQPACKET = 5
    }

    public enum AddressFamily : int
    {
        INET = 2,
        INET6 = 23
    }

    public enum Protocol : int
    {
        IP = 0,           // dummy for IP
        ICMP = 1,           // control message protocol
        IPV4 = IP,            // IP inside IP
        TCP = 6,           // tcp
        UDP = 17,          // user datagram protocol
        IPV6 = 41,          // IPv6 in IPv6
        NONE = 59,          // No next header
        RAW = 255,         // raw IP packet
        MAX = 256,
    }

    public enum SocketOption : int
    {

        /// <summary>
        /// Set to 0 for Non-Blocking TCP connections?  http://e2e.ti.com/support/low_power_rf/f/851/t/269806.aspx
        /// </summary>
        ReceiveBlockingMode = 0,

        /// <summary>
        /// Option to configure recv and recvfrom timeout, value is 10's of milliseconds, 0 for infinite
        /// </summary>
        ReceiveTimeout =  1,           //  
        
        /// <summary>
        /// accept blocking mode: 0 for blocking disabled (default), 1 for blocking enabled
        /// </summary>
        AcceptBlockingMode     =  2           // 

    }
}
